import nmt_chainer.external_libs.bpe.learn_bpe as learn_bpe
import nmt_chainer.external_libs.bpe.apply_bpe as apply_bpe
import collections
import operator
from collections import OrderedDict

from nmt_chainer.dataprocessing.indexer import Indexer

import logging
import json
import codecs
import itertools
import re
import copy

logging.basicConfig()
log = logging.getLogger("rnns:processors")
log.setLevel(logging.INFO)


def build_index_from_iterable(iterable, voc_limit=None):
    counts = collections.defaultdict(int)
    for num_ex, line in enumerate(iterable):
        for w in line:
            counts[w] += 1

    sorted_counts = sorted(
        counts.items(), key=operator.itemgetter(1), reverse=True)

    res = Indexer()

    for w, _ in sorted_counts[:voc_limit]:
        res.add_word(w, should_be_new=True)
    res.finalize()

    return res


class ApplyToMultiIterator(object):
    def __init__(self, iterable, function, can_iter=False):
        self.iterable = iterable
        self.function = function
        if can_iter:
            self.iterator = iter(self.iterable)

    def __iter__(self):
        return ApplyToMultiIterator(self.iterable, self.function, can_iter=True)

    def next(self):
        elem = self.iterator.next()
        return self.function(elem)


class ApplyToMultiIteratorPair(object):
    def __init__(self, iterable1, iterable2, function, can_iter=False):
        self.iterable1 = iterable1
        self.iterable2 = iterable2

        self.function = function
        if can_iter:
            self.iterator1 = iter(self.iterable1)
            self.iterator2 = iter(self.iterable2)

    def __iter__(self):
        return ApplyToMultiIteratorPair(self.iterable1, self.iterable2, self.function, can_iter=True)

    def next(self):
        elem1 = self.iterator1.next()
        elem2 = self.iterator2.next()
        return self.function(elem1, elem2)


class FileMultiIterator(object):
    def __init__(self, filename, max_nb_ex=None, can_iter=False):
        self.filename = filename
        self.max_nb_ex = max_nb_ex

        if can_iter:
            self.f = codecs.open(self.filename, encoding="utf8")
            self.nb_line_read = 0

    def next(self):
        if self.max_nb_ex is not None and self.nb_line_read >= self.max_nb_ex:
            raise StopIteration()
        line = self.f.readline()
        if len(line) == 0:
            raise StopIteration()
        else:
            #             print self.nb_line_read, line.strip()
            self.nb_line_read += 1
            return line.strip()

    def __iter__(self):
        return FileMultiIterator(self.filename, max_nb_ex=self.max_nb_ex, can_iter=True)


REGISTERED_PROCESSORS = []
NAME_TO_PROCESSOR = {}


def register_processor(processor):
    global REGISTERED_PROCESSORS
    global NAME_TO_PROCESSOR
    REGISTERED_PROCESSORS.append(processor)
    assert processor.processor_name() not in NAME_TO_PROCESSOR
    NAME_TO_PROCESSOR[processor.processor_name()] = processor


def get_processor_from_name(name):
    global NAME_TO_PROCESSOR
    return NAME_TO_PROCESSOR[name]


def registered_processor(cls):
    register_processor(cls)
    return cls


class PreProcessor(object):
    def __init__(self):
        self.is_initialized_ = False

    @classmethod
    def processor_name(cls):
        return cls.__name__

    def __str__(self):
        return "<Processor:%s>" % self.processor_name()

#     def make_new_stat(self):
#         return self.Stats()

    def initialize(self, iterable):
        self.is_initialized_ = True

    def is_initialized(self):
        return self.is_initialized_

    @classmethod
    def make_from_serializable(cls, obj):
        assert obj["type"] == "processor"
        processor_name = obj["name"]
        processor = get_processor_from_name(processor_name)
        return processor.make_from_serializable(obj)

    def to_serializable(self):
        raise NotImplemented()

    @classmethod
    def make_base_serializable_object(cls):
        obj = OrderedDict([
            ("type", "processor"),
            ("name", cls.processor_name()),
            ("rev", 1.0)
        ])
        return obj


class MonoProcessor(PreProcessor):
    def convert(self, sentence):
        raise NotImplemented()

    def apply_to_iterable(self, iterable):
        return ApplyToMultiIterator(iterable, lambda elem: self.convert(elem))
#         for sentence in iterable:
#             yield self.convert(sentence, stats = stats)

    def __add__(self, other):
        assert not isinstance(self, ProcessorChain)
        return ProcessorChain([self]) + other


class BiProcessor(PreProcessor):
    def convert(self, sentence1, sentence2):
        raise NotImplemented

    def deconvert(self, sentence1, sentence2):
        raise NotImplemented

    def src_processor(self):
        raise NotImplemented

    def tgt_processor(self):
        raise NotImplemented

    def apply_to_iterable(self, iterable1, iterable2):
        return ApplyToMultiIteratorPair(iterable1, iterable2, lambda elem1, elem2: self.convert(elem1, elem2))


@registered_processor
class ProcessorChain(MonoProcessor):
    def __init__(self, processor_list=[]):
        self.processor_list = processor_list

#     class Stats(object):
#         def __init__(self, stats_list):
#             self.stats_list = stats_list
#
#         def make_report(self):
#             result = []
#             for num, stats in enumerate(self.stats_list):
#                 if stats.__class__ is PreProcessor.Stats:
#                     continue
#                 result.append(stats.make_report())
#             return "\n".join(result)
#
#         def get_sub_stats(self, num):
#             return self.stats_list[num]

    def __str__(self):
        return "[%s]" % (" -> ".join([str(processor) for processor in self.processor_list]))

    def __add__(self, other):
        if isinstance(other, ProcessorChain):
            return ProcessorChain(self.processor_list + other.processor_list)
        else:
            return ProcessorChain(self.processor_list + [other])

    def make_new_stat(self):
        return self.Stats([p.make_new_stat() for p in self.processor_list])

    def initialize(self, iterable):
        for num_processor, processor in enumerate(self.processor_list):
            processor.initialize(iterable)
            if num_processor < len(self.processor_list) - 1:
                iterable = processor.apply_to_iterable(iterable)

    def deconvert(self, seq):
        for num_processor, processor in enumerate(self.processor_list[::-1]):
            seq = processor.deconvert(seq)
        return seq

    def convert(self, sentence, stats=None):
        for num_processor, processor in enumerate(self.processor_list):
            if stats is not None:
                this_stats = stats.get_sub_stats(num_processor)
            else:
                this_stats = None
#             print num_processor, sentence, processor, this_stats, this_stats.make_report()
            sentence = processor.convert(sentence, stats=this_stats)
        return sentence

    def is_initialized(self):
        all_initialized = all(processor.is_initialized()
                              for processor in self.processor_list)
        any_initialized = any(processor.is_initialized()
                              for processor in self.processor_list)
        if all_initialized != any_initialized:
            raise AssertionError(
                repr([processor.is_initialized() for processor in self.processor_list]))
        return all_initialized

    @staticmethod
    def processor_name():
        return "processor_chain"

    @classmethod
    def make_from_serializable(cls, obj):
        res = ProcessorChain()
        res.processor_list = [PreProcessor.make_from_serializable(
            subobj) for subobj in obj["processors_list"]]
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["processors_list"] = [processor.to_serializable()
                                  for processor in self.processor_list]
        return obj


@registered_processor
class BiProcessorChain(BiProcessor):
    def __init__(self):
        self.processors_list = []
        self.src_processor_ = ProcessorChain()
        self.tgt_processor_ = ProcessorChain()

    def __str__(self):
        return "[%s]" % (" -> ".join(["[%s]%s" % (channel, str(processor)) for (channel, processor) in self.processors_list]))

    def is_initialized(self):
        all_initialized = all(processor.is_initialized()
                              for channel, processor in self.processors_list)
        any_initialized = any(processor.is_initialized()
                              for channel, processor in self.processors_list)
        if all_initialized != any_initialized:
            raise AssertionError(repr(
                [processor.is_initialized() for channel, processor in self.processors_list]))
        return all_initialized

    def initialize(self, iterable1, iterable2):
        for num_processor, (channel, processor) in enumerate(
                self.processors_list):
            if channel == "src":
                processor.initialize(iterable1)
                if num_processor < len(self.processors_list) - 1:
                    iterable1 = processor.apply_to_iterable(iterable1)
            elif channel == "tgt":
                processor.initialize(iterable2)
                if num_processor < len(self.processors_list) - 1:
                    iterable2 = processor.apply_to_iterable(iterable2)
            elif channel == "all":
                processor.initialize(iterable1, iterable2)
                if num_processor < len(self.processors_list) - 1:
                    iterable1, iterable2 = processor.apply_to_iterable(iterable1, iterable2)

    def convert(self, sentence1, sentence2):
        for num_processor, (channel, processor) in enumerate(
                self.processors_list):
            if channel == "src":
                sentence1 = processor.convert(sentence1)
            elif channel == "tgt":
                sentence2 = processor.convert(sentence2)
            elif channel == "all":
                sentence1, sentence2 = processor.convert(sentence1, sentence2)
        return sentence1, sentence2

    def deconvert(self, sentence1, sentence2):
        for num_processor, (channel, processor) in enumerate(
                self.processors_list[::-1]):
            if channel == "src":
                sentence1 = processor.deconvert(sentence1)
            elif channel == "tgt":
                sentence2 = processor.deconvert(sentence2)
            elif channel == "all":
                sentence1, sentence2 = processor.deconvert(
                    sentence1, sentence2)
        return sentence1, sentence2

    def src_processor(self):
        return self.src_processor_

    def tgt_processor(self):
        return self.tgt_processor_

    def add_src_processor(self, processor):
        assert isinstance(processor, PreProcessor)
        self.processors_list.append(("src", processor))
        self.src_processor_ += processor

    def add_tgt_processor(self, processor):
        assert isinstance(processor, PreProcessor)
        self.processors_list.append(("tgt", processor))
        self.tgt_processor_ += processor

    def add_biprocessor(self, processor):
        assert isinstance(processor, BiProcessor)
        self.processors_list.append(("all", processor))
        src_processor = processor.src_processor()
        if src_processor is not None:
            self.src_processor_ += src_processor

        tgt_processor = processor.tgt_processor()
        if tgt_processor is not None:
            self.tgt_processor_ += tgt_processor

    @classmethod
    def make_from_serializable(cls, obj):
        res = BiProcessorChain()
        for ch, processor in [[channel, PreProcessor.make_from_serializable(subobj)] for channel, subobj in obj["processors_list"]]:
            if ch == "src":
                res.add_src_processor(processor)
            elif ch == "tgt":
                res.add_tgt_processor(processor)
            else:
                res.add_biprocessor(processor)
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["processors_list"] = [[channel, processor.to_serializable()]
                                  for channel, processor in self.processors_list]
        return obj


@registered_processor
class JointBPEBiProcessor(BiProcessor):
    def __init__(self, bpe_data_file, symbols=10000, min_frequency=2, separator="@@"):
        self.bpe_processor = BPEProcessing(bpe_data_file, symbols=symbols, min_frequency=min_frequency, separator=separator)

    def initialize(self, iterable1, iterable2):
        self.bpe_processor.initialize(itertools.chain(iterable1, iterable2))

    def convert(self, sentence1, sentence2):
        return self.bpe_processor.convert(sentence1), self.bpe_processor.convert(sentence2)

    def deconvert(self, sentence1, sentence2):
        return self.bpe_processor.deconvert(sentence1), self.bpe_processor.deconvert(sentence2)

    def src_processor(self):
        return self.bpe_processor

    def tgt_processor(self):
        return self.bpe_processor

    def is_initialized(self):
        return self.bpe_processor.is_initialized()

    @classmethod
    def make_from_serializable(cls, obj):
        bpe_processor = PreProcessor.make_from_serializable(obj["bpe_processor"])
        res = JointBPEBiProcessor(bpe_data_file="dummy")
        res.bpe_processor = bpe_processor
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["bpe_processor"] = self.bpe_processor.to_serializable()
        return obj


@registered_processor
class BPEProcessing(MonoProcessor):
    def __init__(self, bpe_data_file, symbols=10000, min_frequency=2, separator="@@"):
        self.bpe_data_file = bpe_data_file
        self.symbols = symbols
        self.min_frequency = min_frequency
        self.separator = separator
        self.is_initialized_ = False

    def load_bpe(self):
        log.info("loading BPE data from %s", self.bpe_data_file)
        with codecs.open(self.bpe_data_file, "r", encoding="utf8") as codes:
            self.bpe = apply_bpe.BPE(codes, self.separator)
        self.is_initialized_ = True

    def initialize(self, iterable):
        log.info("Creating BPE data and saving it to %s", self.bpe_data_file)
        with codecs.open(self.bpe_data_file, "w", encoding="utf8") as output:
            learn_bpe.learn_bpe_from_sentence_iterable(iterable, output=output,
                                                       symbols=self.symbols,
                                                       min_frequency=self.min_frequency,
                                                       verbose=False)
        self.load_bpe()

    def __str__(self):
        if self.is_initialized():
            return "bpe<%i-%s>" % (self.symbols, self.bpe_data_file)
        else:
            return "bpe<%i>" % (self.symbols)

    def convert(self, sentence, stats=None):
        assert self.is_initialized()
#         print sentence
#         print "->"
        converted = self.bpe.segment_splitted(sentence)
#         print converted
        return converted

    def deconvert(self, seq):
        res = []
        merge_to_previous = False
        for position, w in enumerate(seq):
            if len(w) >= len(self.separator) and w[-len(self.separator):] == self.separator and position != (len(seq) - 1):
                has_separator = True
                w = w[:-len(self.separator)]
            else:
                has_separator = False

            if merge_to_previous:
                res[-1] = res[-1] + w
            else:
                res.append(w)

            merge_to_previous = has_separator
        return res

    @staticmethod
    def processor_name():
        return "bpe"

    @classmethod
    def make_from_serializable(cls, obj):
        res = BPEProcessing(obj["bpe_data_file"], symbols=obj["symbols"],
                            min_frequency=obj["min_frequency"], separator=obj["separator"])
        res.load_bpe()
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["bpe_data_file"] = self.bpe_data_file
        obj["separator"] = self.separator
        obj["symbols"] = self.symbols
        obj["min_frequency"] = self.min_frequency
        return obj


def segment(line, type="word"):
    if type == "word":
        return line.split(" ")
    elif type == "word2char":
        return tuple("".join(line.split(" ")))
    elif type == "char":
        return tuple(line)
    else:
        raise NotImplemented


@registered_processor
class SimpleSegmenter(MonoProcessor):
    def __init__(self, type="word"):
        assert type in "word char word2char".split()
        self.type = type
        self.is_initialized_ = False

    def convert(self, sentence, stats=None):
        return segment(sentence, type=self.type)

    def deconvert(self, seq):
        if self.type == "word":
            return " ".join(seq)
        elif self.type == "char":
            return "".join(seq)
        elif self.type == "word2char":
            return "".join(seq)
        else:
            raise ValueError()

    @staticmethod
    def processor_name():
        return "simple_segmenter"

    def __str__(self):
        return "segmenter<%s>" % self.type

    @classmethod
    def make_from_serializable(cls, obj):
        res = SimpleSegmenter()
        res.type = obj["segmentation_type"]
        res.is_initialized_ = True
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["segmentation_type"] = self.type
        return obj


@registered_processor
class LatinScriptProcess(MonoProcessor):

    CAP_CHAR = u"\u203B"
    ALL_CAPS_CHAR = u"\u203C"
    SUFFIX_CHAR = u"\u203F"

    def __init__(self, mode="all_adjoint"):
        assert mode in "all_adjoint caps_isolate".split()
        self.mode = mode
        self.is_initialized_ = False

    def convert_caps(self, w):
        if self.CAP_CHAR in w or self.ALL_CAPS_CHAR in w:
            raise ValueError("Special char in word")
        assert len(w) > 0
        if w.istitle():
            return self.CAP_CHAR + w.lower()
        elif w.isupper():
            return self.ALL_CAPS_CHAR + w.lower()
        else:
            return w

    def deconvert_caps(self, w):
        if w.startswith(self.CAP_CHAR):
            assert not w[1:].istitle()
            return w[1:].title()  # w[1].upper() + w[2:]
        elif w.startswith(self.ALL_CAPS_CHAR):
            #             assert w[1:].islower()
            return w[1:].upper()
        else:
            return w

    def convert_caps_alt(self, w):
        if self.CAP_CHAR in w or self.ALL_CAPS_CHAR in w or " " in w:
            raise ValueError("Special char in word")
        assert len(w) > 0
        if w.istitle():
            return self.CAP_CHAR + " " + w.lower()
        elif w.isupper():
            return self.ALL_CAPS_CHAR + " " + w.lower()
        else:
            return w

    def deconvert_caps_alt_sentence(self, sentence):
        res = []
        next_is_cap = False
        next_is_all_caps = False
        for w in sentence.split(" "):
            if w == self.CAP_CHAR:
                next_is_cap = True
                continue
            if w == self.ALL_CAPS_CHAR:
                next_is_all_caps = True
                continue
            if next_is_all_caps:
                w = w.upper()
            elif next_is_cap:
                w = w.title()
            next_is_cap = False
            next_is_all_caps = False
            res.append(w)
        return " ".join(res)

    def convert_punct_word(self, w):
        if self.SUFFIX_CHAR in w:
            raise ValueError("Special char in word")
        if len(w) > 3 and w.endswith("..."):
            w = w[:-3] + " " + self.SUFFIX_CHAR + "..."
        else:
            if len(w) > 1:
                for punct in ".!?:,;\"%$'`)]":
                    if w.endswith(punct):
                        w = w[:-1] + " " + self.SUFFIX_CHAR + punct
                        break
        return w

    def convert_punct_inside(self, w):
        if self.SUFFIX_CHAR in w:
            raise ValueError("Special char in word")

        if w == "...":
            return w
        ends_with_triple_dots = False
        if len(w) > 3 and w.endswith("..."):
            ends_with_triple_dots = True
            w = w[:-3]

        if len(w) > 1:
            for punct in ".!?:,;\"%$'`)]-([&=+*<>_/\\^~#@|":
                splitted = w.split(punct)
                if len(splitted[0]) == 0:
                    splitted = [punct + splitted[1]] + splitted[2:]  # do not split if punc at the beginning
                w = (" " + self.SUFFIX_CHAR + punct).join(splitted)
        if ends_with_triple_dots:
            w = w + " " + self.SUFFIX_CHAR + "..."
        return w

    def deconvert_punct_sentence(self, sentence):
        res = []
        for w in sentence.split(" "):
            if len(res) > 0 and w.startswith(self.SUFFIX_CHAR):
                res[-1] = res[-1] + w[1:]
            else:
                res.append(w)
        return " ".join(res)

    def convert(self, sentence, stats=None):
        sentence = re.sub("\s+", " ", sentence)
        sentence = " ".join(self.convert_punct_inside(w)
                            for w in sentence.split(" "))
        if self.mode == "all_adjoint":
            sentence = " ".join(self.convert_caps(w)
                                for w in sentence.split(" "))
        elif self.mode == "caps_isolate":
            sentence = " ".join(self.convert_caps_alt(w)
                                for w in sentence.split(" "))
        return sentence

    def deconvert(self, sentence):
        if self.mode == "all_adjoint":
            sentence = " ".join(self.deconvert_caps(w)
                                for w in sentence.split(" "))
        elif self.mode == "caps_isolate":
            sentence = self.deconvert_caps_alt_sentence(sentence)
        sentence = self.deconvert_punct_sentence(sentence)
        return sentence

    @staticmethod
    def processor_name():
        return "latin_script_processor"

    def __str__(self):
        return "latin_script_processor"

    @classmethod
    def make_from_serializable(cls, obj):
        res = LatinScriptProcess()
        res.mode = obj.get("mode", "all_adjoint")
        res.is_initialized_ = True
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["mode"] = self.mode
        return obj


#     class IterableIterator(object):
#         def __init__(self, filename):
#             self.filename = filename
#         def make_iter(self):
#
#
#     def __iter__(self):
#         with codecs.open(self.filename, encoding="utf8") as f:
#             for num_line, line in enumerate(f):
# #                 print self.filename, num_line, self.max_nb_ex
#                 if self.max_nb_ex is not None and num_line >= self.max_nb_ex:
#                     return
#                 yield line.strip()

class IndexingPrePostProcessorBase(MonoProcessor):
    def __len__(self):
        raise NotImplemented

    def is_unk_idx(self, idx):
        raise NotImplemented

    def convert(self, sentence, stats=None):
        raise NotImplemented

    def convert_swallow(self, sentence, stats=None):
        raise NotImplemented

    def deconvert_swallow(self, seq, unk_tag="#UNK#", no_oov=True, eos_idx=None):
        raise NotImplemented

    def deconvert_post(self, seq):
        raise NotImplemented

    def deconvert(self, seq, unk_tag="#UNK#", no_oov=True, eos_idx=None):
        sentence = self.deconvert_swallow(
            seq, unk_tag=unk_tag, no_oov=no_oov, eos_idx=eos_idx)
        sentence = self.deconvert_post(sentence)
        return sentence

    class Stats(object):
        def make_report(self):
            return "nothing to report"

        def report_as_obj(self):
            return OrderedDict()

    def make_new_stat(self):
        return self.Stats()


@registered_processor
class BiIndexingPrePostProcessor(BiProcessor):
    def __init__(self, voc_limit1=None, voc_limit2=None):
        self.indexer1 = IndexingPrePostProcessor(voc_limit1)
        self.indexer2 = IndexingPrePostProcessor(voc_limit2)
        self.preprocessor = None
        self.is_initialized_ = False

    def convert(self, sentence1, sentence2, stat1=None, stat2=None):
        if self.preprocessor is not None:
            sentence1, sentence2 = self.preprocessor.convert(sentence1, sentence2)
        return self.indexer1.convert_swallow(sentence1, stat1), self.indexer2.convert_swallow(sentence2, stat2)

    def deconvert(self, sentence1, sentence2):
        sentence1, sentence2 = self.indexer1.deconvert_swallow(sentence1), self.indexer2.deconvert_swallow(sentence2)
        if self.preprocessor is not None:
            sentence1, sentence2 = self.preprocessor.deconvert(
                sentence1, sentence2)
        return sentence1, sentence2

    def __str__(self):
        return """BiIndexing:
    src: %s tgt: %s
    pp: %s""" % (self.indexer1, self.indexer2, self.preprocessor)

    def src_processor(self):
        return self.indexer1

    def tgt_processor(self):
        return self.indexer2

    def make_new_stat(self):
        return self.indexer1.make_new_stat(), self.indexer2.make_new_stat()

    def initialize(self, iterable1, iterable2):
        if self.preprocessor is not None:
            self.preprocessor.initialize(iterable1, iterable2)
            iterable_1_2 = self.preprocessor.apply_to_iterable(
                iterable1, iterable2)
            iterable1 = ApplyToMultiIterator(iterable_1_2, lambda x: x[0])
            iterable2 = ApplyToMultiIterator(iterable_1_2, lambda x: x[1])

        self.indexer1.initialize_swallow(iterable1)
        self.indexer2.initialize_swallow(iterable2)
        self.is_initialized_ = True

    def add_preprocessor(self, processor, can_be_initialized=False):
        if not can_be_initialized and self.is_initialized():
            raise ValueError("adding preprocessor to initialized biindexer")
        self.preprocessor = processor
        self.indexer1.add_preprocessor(processor.src_processor(), can_be_initialized=can_be_initialized)
        self.indexer2.add_preprocessor(processor.tgt_processor(), can_be_initialized=can_be_initialized)

    @classmethod
    def make_from_serializable(cls, obj):
        if isinstance(obj, list):
            if len(obj) != 2:
                raise ValueError("unrecognized preprocessor format")
            res = BiIndexingPrePostProcessor()
            obj_src, obj_tgt = obj
            indexer1 = IndexingPrePostProcessor.make_from_serializable(obj_src)
            indexer2 = IndexingPrePostProcessor.make_from_serializable(obj_tgt)

            res.indexer1 = indexer1
            res.indexer2 = indexer2

            if indexer1.preprocessor is not None or indexer2.preprocessor is not None:
                preprocessor = BiProcessorChain()
                if indexer1.preprocessor is not None:
                    preprocessor.add_src_processor(indexer1.preprocessor)
                    indexer1.preprocessor = None
                if indexer2.preprocessor is not None:
                    preprocessor.add_tgt_processor(indexer2.preprocessor)
                    indexer2.preprocessor = None
                res.add_preprocessor(preprocessor, can_be_initialized=True)
        else:
            res = BiIndexingPrePostProcessor()
            res.indexer1 = PreProcessor.make_from_serializable(
                obj["indexer_src"])
            res.indexer2 = PreProcessor.make_from_serializable(
                obj["indexer_tgt"])
            assert isinstance(res.indexer1, IndexingPrePostProcessorBase)
            assert isinstance(res.indexer2, IndexingPrePostProcessorBase)
            if "preprocessor" in obj:
                preprocessor = PreProcessor.make_from_serializable(obj["preprocessor"])
                res.add_preprocessor(preprocessor, can_be_initialized=True)
            else:
                res.preprocessor = None
            res.is_initialized_ = True
        return res

    def to_serializable(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["indexer_src"] = self.indexer1.to_serializable_swallow()
        obj["indexer_tgt"] = self.indexer2.to_serializable_swallow()
        if self.preprocessor is not None:
            obj["preprocessor"] = self.preprocessor.to_serializable()
        return obj


@registered_processor
class IndexingPrePostProcessor(IndexingPrePostProcessorBase):
    def __init__(self, voc_limit=None):
        self.voc_limit = voc_limit
#         self.save_index_to = save_index_to
        self.is_initialized_ = False
        self.preprocessor = None

    class Stats(object):
        def __init__(self):
            self.unk_cnt = 0
            self.token = 0
            self.nb_ex = 0

        def update(self, unk_cnt, token, nb_ex):
            #             print unk_cnt, token, nb_ex, self.unk_cnt, self.token, self.nb_ex
            self.unk_cnt += unk_cnt
            self.token += token
            self.nb_ex += nb_ex

        def make_report(self):
            report = "#tokens: %i   of which %i (%f%%) are unknown" % (self.token,
                                                                       self.unk_cnt,
                                                                       (self.unk_cnt * 100.0) / self.token if self.token != 0 else 0
                                                                       )
            return report

        def report_as_obj(self):
            return OrderedDict([
                ("nb_tokens", self.token),
                ("nb_unk", self.unk_cnt),
                ("unknown_percent", (self.unk_cnt * 100.0) / self.token if self.token != 0 else 0)
            ])

    def initialize_swallow(self, iterable):
        log.info("building dic")
        self.indexer = build_index_from_iterable(iterable, self.voc_limit)
        self.is_initialized_ = True

    def initialize(self, iterable):
        if self.preprocessor is not None:
            self.preprocessor.initialize(iterable)
            iterable = self.preprocessor.apply_to_iterable(iterable)
        self.initialize_swallow(iterable)
        self.indexer = build_index_from_iterable(iterable, self.voc_limit)

    def __str__(self):
        if not self.is_initialized():
            return "indexer<%i>pp[%s]" % (self.voc_limit, "" if self.preprocessor is None else str(self.preprocessor))
        else:
            return "indexer<%i - %i>pp[%s]" % (self.voc_limit, len(self.indexer), "" if self.preprocessor is None else str(self.preprocessor))

    def __len__(self):
        if not self.is_initialized():
            return 0
        else:
            return len(self.indexer)

    def is_initialized(self):
        return self.is_initialized_

    def convert(self, sentence, stats=None):
        if self.preprocessor is not None:
            sentence = self.preprocessor.convert(sentence)
        converted = self.convert_swallow(sentence, stats=stats)
        return converted

    def convert_swallow(self, sentence, stats=None):
        assert self.is_initialized()
        converted = self.indexer.convert(sentence)
        if stats is not None:
            unk_cnt = sum(self.indexer.is_unk_idx(w) for w in converted)
            stats.update(unk_cnt=unk_cnt, token=len(converted), nb_ex=1)
        return converted

    def apply_to_iterable(self, iterable, stats=None):
        raise AssertionError()

    def deconvert_swallow(self, seq, unk_tag="#UNK#", no_oov=True, eos_idx=None):
        return self.indexer.deconvert(seq, unk_tag=unk_tag, no_oov=no_oov, eos_idx=eos_idx)

    def deconvert_post(self, seq):
        if self.preprocessor is not None:
            seq = self.preprocessor.deconvert(seq)
        return seq

    def is_unk_idx(self, idx):
        return self.indexer.is_unk_idx(idx)

    @staticmethod
    def processor_name():
        return "indexer"

    @classmethod
    def make_serialized_form_compatible_with_newer_version(cls, obj):
        if Indexer.check_if_data_indexer(obj):
            new_obj = cls.make_base_serializable_object()
            new_obj["indexer"] = obj
            new_obj["voc_limit"] = len(obj)
            ss = SimpleSegmenter("word")
            ss.initialize(None)
            new_obj["preprocessor"] = ss.to_serializable()

        elif "processors_list" in obj:
            new_obj = obj["processors_list"][-1]
            if len(obj["processors_list"]) > 1:
                preproc_obj = copy.deepcopy(obj)
                preproc_obj["processors_list"] = preproc_obj["processors_list"][:-1]
                new_obj["preprocessor"] = preproc_obj
        else:
            new_obj = obj
        return new_obj

    @classmethod
    def make_from_serializable(cls, obj):
        obj = cls.make_serialized_form_compatible_with_newer_version(obj)
        res = IndexingPrePostProcessor(voc_limit=obj["voc_limit"])
        res.indexer = Indexer.make_from_serializable(obj["indexer"])
        if "preprocessor" in obj:
            res.preprocessor = PreProcessor.make_from_serializable(
                obj["preprocessor"])
        res.is_initialized_ = True
        return res

    def to_serializable(self):
        obj = self.to_serializable_swallow()
        if self.preprocessor is not None:
            obj["preprocessor"] = self.preprocessor.to_serializable()
        return obj

    def to_serializable_swallow(self):
        assert self.is_initialized()
        obj = self.make_base_serializable_object()
        obj["voc_limit"] = self.voc_limit
#         obj["voc_fn"] = self.save_index_to
        obj["indexer"] = self.indexer.to_serializable()
        return obj

    @classmethod
    def make_from_indexer(cls, indexer):
        res = IndexingPrePostProcessor(voc_limit=len(indexer))
        res.indexer = indexer
        res.is_initialized_ = True
        return res

    def add_preprocessor(self, processor, can_be_initialized=False):
        if not can_be_initialized and self.is_initialized():
            raise ValueError("adding procesor to initialized indexer")
        self.preprocessor = processor


def izip_must_equal(it1, it2):
    for s1, s2 in itertools.izip_longest(it1, it2, fillvalue=None):
        if s1 is None or s2 is None:
            raise ValueError("iterators have different sizes")
        yield s1, s2


def build_dataset_pp(src_fn, tgt_fn, bi_idx, max_nb_ex=None):
    #                   src_voc_limit=None, tgt_voc_limit=None, max_nb_ex=None, dic_src=None, dic_tgt=None,
    #                   tgt_segmentation_type="word", src_segmentation_type="word"):

    src = FileMultiIterator(src_fn, max_nb_ex=max_nb_ex)
    tgt = FileMultiIterator(tgt_fn, max_nb_ex=max_nb_ex)

    if not bi_idx.is_initialized():
        bi_idx.initialize(src, tgt)
#
#     if not src_pp.is_initialized():
#         log.info("building src_dic")
#         src_pp.initialize(src)
#
#     if not tgt_pp.is_initialized():
#         log.info("building tgt_dic")
#         tgt_pp.initialize(tgt)
#
#     print src_pp
#
#     print tgt_pp

    print bi_idx

    stats_src, stats_tgt = bi_idx.make_new_stat()

    log.info("start indexing")

    res = []

    for sentence_src, sentence_tgt in izip_must_equal(src, tgt):
        #         print len(sentence_tgt), len(sentence_src)
        seq_src, seq_tgt = bi_idx.convert(sentence_src, sentence_tgt, stats_src, stats_tgt)
        res.append((seq_src, seq_tgt))

    return res, stats_src, stats_tgt


def build_dataset_one_side_pp(src_fn, src_pp, max_nb_ex=None):

    src = FileMultiIterator(src_fn, max_nb_ex=max_nb_ex)
    print src_pp
    if not src_pp.is_initialized():
        log.info("building src_dic")
        src_pp.initialize(src)

    stats_src = src_pp.make_new_stat()
    log.info("start indexing")

    res = []

    for sentence_src in src:
        #         print len(sentence_tgt), len(sentence_src)
        seq_src = src_pp.convert(sentence_src, stats=stats_src)
        res.append(seq_src)

    return res, stats_src


# def load_pp_from_data(data):
#     if Indexer.check_if_data_indexer(data):
#         indexer = Indexer.make_from_serializable(data)
#         ipp = IndexingPrePostProcessor.make_from_indexer(indexer)
#         ss_pp = SimpleSegmenter("word")
#         ss_pp.initialize(None)
#         pp = ProcessorChain([ss_pp,
#                                  ipp])
#     else:
#         pp = PreProcessor.make_from_serializable(data)
#
#     return pp

def load_pp_pair_from_file(filename):
    bi_idx = BiIndexingPrePostProcessor.make_from_serializable(
        json.load(open(filename)))
    log.info("loading bilingual preprocessor from file %s", filename)
    log.info(str(bi_idx))
#     print bi_idx.src_processor()
#     print bi_idx.tgt_processor()
    return bi_idx


def save_pp_pair_to_file(bi_idx, filename):
    json.dump(bi_idx.to_serializable(), open(filename, "w"), indent=2, separators=(',', ': '))



#!/usr/bin/env python
"""training.py: training procedures."""
from nmt_chainer.utilities import argument_parsing_tools
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

from chainer import serializers
import time

import logging
import sys
# import h5py

import math
import json

from nmt_chainer.utilities.utils import minibatch_provider, minibatch_provider_curiculum, make_batch_src_tgt
from nmt_chainer.translation.evaluation import (
    compute_loss_all, translate_to_file, sample_once)

import chainer.iterators
import chainer.dataset.iterator
import chainer.training
import chainer.training.extensions

logging.basicConfig()
log = logging.getLogger("rnns:training")
log.setLevel(logging.INFO)


def sent_complexity(sent):
    rank_least_common_word = max(sent)
    length = len(sent)
    return length * math.log(rank_least_common_word + 1)


def example_complexity(ex):
    return sent_complexity(ex[0]) + sent_complexity(ex[1])


import numpy


class SerialIteratorWithPeek(chainer.iterators.SerialIterator):

    def peek(self):
        """
        Return the next batch of data without updating its internal state.
        Several call to peek() should return the same result. A call to next()
        after a call to peek() will return the same result as the previous peek.
        """
        if not self._repeat and self.epoch > 0:
            raise StopIteration

        i = self.current_position
        i_end = i + self.batch_size
        N = len(self.dataset)
        if self._order is None:
            batch = self.dataset[i:i_end]
        else:
            batch = [self.dataset[index] for index in self._order[i:i_end]]

        if i_end >= N:
            if self._repeat:
                rest = i_end - N
                if self._order is not None:
                    numpy.random.shuffle(self._order)
                if rest > 0:
                    if self._order is None:
                        batch += list(self.dataset[:rest])
                    else:
                        batch += [self.dataset[index]
                                  for index in self._order[:rest]]
        return batch


class LengthBasedSerialIterator(chainer.dataset.iterator.Iterator):
    """
    This iterator will try to return batches with sequences of similar length.
    This is done by first extracting nb_of_batch_to_sort x batch_size sequences, sorting them by size,
    and then successively yielding nb_of_batch_to_sort batches of size batch_size.
    """

    def __init__(self, dataset, batch_size, nb_of_batch_to_sort=20, sort_key=lambda x: len(x[1]),
                 repeat=True, shuffle=True):

        self.sub_iterator = SerialIteratorWithPeek(dataset, batch_size * nb_of_batch_to_sort,
                                                   repeat=repeat, shuffle=shuffle)
        self.dataset = dataset
        self.index_in_sub_batch = 0
        self.sub_batch = None
        self.sort_key = sort_key
        self.batch_size = batch_size
        self.nb_of_batch_to_sort = nb_of_batch_to_sort
        self.repeat = repeat

    def update_sub_batch(self):
        self.sub_batch = list(self.sub_iterator.next())  # copy the result so that we can sort without side effects
        if self.repeat and len(self.sub_batch) != self.batch_size * self.nb_of_batch_to_sort:
            raise AssertionError
        self.sub_batch.sort(key=self.sort_key)
        self.index_in_sub_batch = 0

    def __next__(self):
        if self.sub_batch is None or self.index_in_sub_batch >= self.nb_of_batch_to_sort:
            assert self.sub_batch is None or self.index_in_sub_batch == self.nb_of_batch_to_sort
            self.update_sub_batch()

        minibatch = self.sub_batch[self.index_in_sub_batch * self.batch_size: (self.index_in_sub_batch + 1) * self.batch_size]

        self.index_in_sub_batch += 1

        return minibatch

    def peek(self):
        if self.sub_batch is None or self.index_in_sub_batch >= self.nb_of_batch_to_sort:
            assert self.sub_batch is None or self.index_in_sub_batch == self.nb_of_batch_to_sort
            sub_batch = list(self.sub_iterator.peek())  # copy the result so that we can sort without side effects
            sub_batch.sort(key=self.sort_key)
            index_in_sub_batch = 0
        else:
            sub_batch = self.sub_batch
            index_in_sub_batch = self.index_in_sub_batch
        minibatch = sub_batch[index_in_sub_batch * self.batch_size: (index_in_sub_batch + 1) * self.batch_size]
        return minibatch

    next = __next__

    # It is a bit complicated to keep an accurate value for epoch detail. In practice, the beginning of a sub_batch crossing epoch will have its
    # epoch_detail clipped to epoch
    @property
    def epoch_detail(self):
        remaining_sub_batch_lenth = (self.nb_of_batch_to_sort - self.index_in_sub_batch) * self.batch_size
        assert remaining_sub_batch_lenth >= 0
        epoch_discount = remaining_sub_batch_lenth / float(len(self.dataset))
        sub_epoch_detail = self.sub_iterator.epoch_detail
        epoch_detail = sub_epoch_detail - epoch_discount
        epoch_detail = max(epoch_detail, self.epoch)
        return epoch_detail

    # epoch and is_new_epoch are updated as soon as the end of the current
    # sub_batch has reached a new epoch.
    @property
    def epoch(self):
        return self.sub_iterator.epoch

    @property
    def is_new_epoch(self):
        if self.sub_iterator.is_new_epoch:
            assert self.sub_batch is None or self.index_in_sub_batch > 0
            if self.index_in_sub_batch == 1:
                return True
        return False

    # We do not serialize index_in_sub_batch. A deserialized iterator will start from the next sub_batch.
#     def serialize(self, serializer):
#         self.sub_iterator = serializer("sub_iterator", self.sub_iterator)
#         self.sub_iterator.serialize(serializer["sub_iterator"])


import six


def make_collection_of_variables(in_arrays, volatile="off"):
    if isinstance(in_arrays, tuple):
        in_vars = tuple(chainer.variable.Variable(x, volatile=volatile) for x in in_arrays)
    elif isinstance(in_arrays, dict):
        in_vars = {key: chainer.variable.Variable(x, volatile=volatile)
                   for key, x in six.iteritems(in_arrays)}
    else:
        in_vars = chainer.variable.Variable(in_arrays, volatile=volatile)
    return in_vars


class Updater(chainer.training.StandardUpdater):
    def __init__(self, iterator, optimizer, converter=chainer.dataset.convert.concat_examples,
                 device=None, loss_func=None, need_to_convert_to_variables=True):
        super(Updater, self).__init__(iterator, optimizer, converter=converter,
                                      device=device, loss_func=loss_func)
        self.need_to_convert_to_variables = need_to_convert_to_variables

    def update_core(self):
        t0 = time.clock()

        batch = self._iterators['main'].next()
        in_arrays = self.converter(batch, self.device)

        optimizer = self._optimizers['main']
        loss_func = self.loss_func or optimizer.target

        if self.need_to_convert_to_variables:
            in_arrays = make_collection_of_variables(in_arrays)

        t1 = time.clock()

        if isinstance(in_arrays, tuple):
            optimizer.update(loss_func, *in_arrays)
        elif isinstance(in_arrays, dict):
            optimizer.update(loss_func, **in_arrays)
        else:
            optimizer.update(loss_func, in_arrays)

        t2 = time.clock()
        update_duration = t2 - t0
        mb_preparation_duration = t1 - t0
        optimizer_update_cycle_duration = t2 - t1
        chainer.reporter.report({"update_duration": update_duration,
                                 "mb_preparation_duration": mb_preparation_duration,
                                 "optimizer_update_cycle_duration": optimizer_update_cycle_duration})


class ComputeLossExtension(chainer.training.Extension):
    priority = chainer.training.PRIORITY_WRITER

    def __init__(self, data, eos_idx,
                 mb_size, gpu, reverse_src, reverse_tgt,
                 save_best_model_to=None, observation_name="dev_loss", config_training=None):
        self.best_loss = None
        self.save_best_model_to = save_best_model_to
        self.observation_name = observation_name
        self.data = data
        self.eos_idx = eos_idx
        self.mb_size = mb_size
        self.gpu = gpu
        self.reverse_src = reverse_src
        self.reverse_tgt = reverse_tgt
        self.config_training = config_training

    def __call__(self, trainer):
        encdec = trainer.updater.get_optimizer("main").target
        log.info("computing %s" % self.observation_name)
        dev_loss = compute_loss_all(encdec, self.data, self.eos_idx, self.mb_size,
                                    gpu=self.gpu,
                                    reverse_src=self.reverse_src, reverse_tgt=self.reverse_tgt)
        log.info("%s: %f (current best: %r)" % (self.observation_name, dev_loss, self.best_loss))
        chainer.reporter.report({self.observation_name: dev_loss})

        if self.best_loss is None or self.best_loss > dev_loss:
            log.info("loss (%s) improvement: %r -> %r" % (self.observation_name,
                                                          self.best_loss, dev_loss))
            self.best_loss = dev_loss
            if self.save_best_model_to is not None:
                log.info("saving best loss (%s) model to %s" % (self.observation_name, self.save_best_model_to,))
                serializers.save_npz(self.save_best_model_to, encdec)
                if self.config_training is not None:
                    config_session = self.config_training.copy(readonly=False)
                    config_session.add_section("model_parameters", keep_at_bottom="metadata")
                    config_session["model_parameters"]["filename"] = self.save_best_model_to
                    config_session["model_parameters"]["type"] = "model"
                    config_session["model_parameters"]["description"] = "best_loss"
                    config_session["model_parameters"]["infos"] = argument_parsing_tools.OrderedNamespace()
                    config_session["model_parameters"]["infos"]["loss"] = float(dev_loss)
                    config_session["model_parameters"]["infos"]["iteration"] = trainer.updater.iteration
                    config_session.set_metadata_modified_time()
                    config_session.save_to(self.save_best_model_to + ".config")
#                     json.dump(config_session, open(self.save_best_model_to + ".config", "w"), indent=2, separators=(',', ': '))

    def serialize(self, serializer):
        self.best_loss = serializer("best_loss", self.best_loss)
        # Make sure that best_loss is at the right location.
        # After deserialization, the best_loss is
        # instanciated on the CPU instead of the GPU.
        if self.gpu is None:
            pass  # best_loss should be on the cpu memory anyway
#             if isinstance(self.best_loss, cupy.core.ndarray):
#                 self.best_loss = cupy.asnumpy(self.best_loss)
        else:
            import cupy
            if self.best_loss is not None and (isinstance(self.best_loss, numpy.ndarray) or self.best_loss.device.id != self.gpu):
                with cupy.cuda.Device(self.gpu):
                    self.best_loss = cupy.array(self.best_loss)


class ComputeBleuExtension(chainer.training.Extension):
    priority = chainer.training.PRIORITY_WRITER

    def __init__(self, data, eos_idx, src_indexer, tgt_indexer,
                 translations_fn, control_src_fn,
                 mb_size, gpu, reverse_src=False, reverse_tgt=False,
                 save_best_model_to=None, observation_name="dev_bleu",
                 nb_steps=50,
                 s_unk_tag=lambda num, utag: "S_UNK_%i" % utag,
                 t_unk_tag=lambda num, utag: "T_UNK_%i" % utag,
                 config_training=None):
        self.best_bleu = None
        self.save_best_model_to = save_best_model_to
        self.observation_name = observation_name
        self.data = data
        self.eos_idx = eos_idx
        self.mb_size = mb_size
        self.gpu = gpu
        self.reverse_src = reverse_src
        self.reverse_tgt = reverse_tgt
        self.s_unk_tag = s_unk_tag
        self.t_unk_tag = t_unk_tag

        self.src_indexer = src_indexer
        self.tgt_indexer = tgt_indexer
        self.nb_steps = nb_steps

        self.translations_fn = translations_fn
        self.control_src_fn = control_src_fn

        self.src_data = [x for x, y in data]
        self.references = [y for x, y in data]

        self.config_training = config_training

    def __call__(self, trainer):
        encdec = trainer.updater.get_optimizer("main").target
#         translations_fn = output_files_dict["dev_translation_output"] #save_prefix + ".test.out"
#         control_src_fn = output_files_dict["dev_src_output"] #save_prefix + ".test.src.out"
        bleu_stats = translate_to_file(encdec, self.eos_idx, self.src_data, self.mb_size,
                                       self.tgt_indexer,
                                       self.translations_fn, test_references=self.references,
                                       control_src_fn=self.control_src_fn,
                                       src_indexer=self.src_indexer, gpu=self.gpu, nb_steps=50,
                                       reverse_src=self.reverse_src, reverse_tgt=self.reverse_tgt,
                                       s_unk_tag=self.s_unk_tag, t_unk_tag=self.t_unk_tag)
        bleu = bleu_stats.bleu()
        chainer.reporter.report({self.observation_name: bleu,
                                 self.observation_name + "_details": repr(bleu)})

        if self.best_bleu is None or self.best_bleu < bleu:
            log.info("%s improvement: %r -> %r" % (self.observation_name, self.best_bleu, bleu))
            self.best_bleu = bleu
            if self.save_best_model_to is not None:
                log.info("saving best bleu (%s) model to %s" % (self.observation_name, self.save_best_model_to,))
                serializers.save_npz(self.save_best_model_to, encdec)
                if self.config_training is not None:
                    config_session = self.config_training.copy(readonly=False)
                    config_session.add_section("model_parameters", keep_at_bottom="metadata")
                    config_session["model_parameters"]["filename"] = self.save_best_model_to
                    config_session["model_parameters"]["type"] = "model"
                    config_session["model_parameters"]["description"] = "best_bleu"
                    config_session["model_parameters"]["infos"] = argument_parsing_tools.OrderedNamespace()
                    config_session["model_parameters"]["infos"]["bleu_stats"] = str(bleu_stats)
                    config_session["model_parameters"]["infos"]["iteration"] = trainer.updater.iteration
                    config_session.set_metadata_modified_time()
                    config_session.save_to(self.save_best_model_to + ".config")
# json.dump(config_session, open(self.save_best_model_to + ".config",
# "w"), indent=2, separators=(',', ': '))
        else:
            log.info("no bleu (%s) improvement: %f >= %f" % (self.observation_name, self.best_bleu, bleu))

    def serialize(self, serializer):
        self.best_bleu = serializer("best_bleu", self.best_bleu)


class TrainingLossSummaryExtension(chainer.training.Extension):
    priority = chainer.training.PRIORITY_EDITOR

    def __init__(self, trigger=(200, 'iteration')):
        self.update_trigger = chainer.training.trigger.get_trigger(trigger)
        self.reset()
#         self.previous_time = None

    def reset(self):
        self.total_loss = 0
        self.total_nb_predictions = 0
        self.total_update_time = 0
        self.nb_observations = 0

    def __call__(self, trainer):
        # accumulate the observations

        mb_avg_loss = float(trainer.observation["mb_loss"]) / trainer.observation["mb_nb_predictions"]
        log.info("E:%i I:%i L:%f U: %.4f = %.4f + %.4f F:%.4f" % (trainer.updater.epoch,
                 trainer.updater.iteration, mb_avg_loss,
                 trainer.observation["update_duration"],
                 trainer.observation["mb_preparation_duration"],
                 trainer.observation["optimizer_update_cycle_duration"],
                 trainer.observation["forward_time"]))

        self.total_loss += trainer.observation["mb_loss"]
        self.total_nb_predictions += trainer.observation["mb_nb_predictions"]
        self.total_update_time += trainer.observation["update_duration"]
        self.nb_observations += 1

        if self.update_trigger(trainer):
            # output the result
            avg_loss = float(self.total_loss) / self.total_nb_predictions
            avg_update_time = self.total_update_time / self.nb_observations
            chainer.reporter.report({"avg_training_loss": avg_loss})
            chainer.reporter.report({"avg_update_time": avg_update_time})
            self.reset()


class SqliteLogExtension(chainer.training.Extension):
    priority = chainer.training.PRIORITY_READER

    def __init__(self, db_path):
        self.db_path = db_path

    def __call__(self, trainer):
        if any((key in trainer.observation)
                for key in "dev_bleu dev_loss test_bleu test_loss avg_training_loss".split()):

            log.info("saving dev results to %s" % (self.db_path))

            import sqlite3
            import datetime
            db_connection = sqlite3.connect(self.db_path)
            db_cursor = db_connection.cursor()
            db_cursor.execute('''CREATE TABLE IF NOT EXISTS exp_data
(date text, bleu_info text, iteration real,
loss real, bleu real,
dev_loss real, dev_bleu real,
valid_loss real, valid_bleu real,
avg_time real, avg_training_loss real)''')

            dev_loss = trainer.observation.get("dev_loss", None)
            if dev_loss is not None:
                dev_loss = float(dev_loss)

            test_loss = trainer.observation.get("test_loss", None)
            if test_loss is not None:
                test_loss = float(test_loss)

            avg_training_loss = trainer.observation.get("avg_training_loss", None)
            if avg_training_loss is not None:
                avg_training_loss = float(avg_training_loss)

            infos = (datetime.datetime.now().strftime("%I:%M%p %B %d, %Y"),
                     trainer.observation.get("test_bleu_details", None), trainer.updater.iteration,
                     test_loss,
                     trainer.observation.get("test_bleu", None),
                     dev_loss,
                     trainer.observation.get("dev_bleu", None),
                     None, None,
                     trainer.observation.get("avg_update_time", None), avg_training_loss)
            db_cursor.execute("INSERT INTO exp_data VALUES (?,?,?,?,?,?,?,?,?,?,?)", infos)
            db_connection.commit()
            db_connection.close()


class CheckpontSavingExtension(chainer.training.Extension):
    priority = chainer.training.PRIORITY_READER

    def __init__(self, save_to, config_training):
        self.save_to = save_to
        self.config_training = config_training

    def __call__(self, trainer):
        log.info("Saving current trainer state to file %s" % self.save_to)
        serializers.save_npz(self.save_to, trainer)
        config_session = self.config_training.copy(readonly=False)
        config_session.add_section("model_parameters", keep_at_bottom="metadata")
        config_session["model_parameters"]["filename"] = self.save_to
        config_session["model_parameters"]["type"] = "snapshot"
        config_session["model_parameters"]["description"] = "checkpoint"
        config_session["model_parameters"]["infos"] = argument_parsing_tools.OrderedNamespace()
        config_session["model_parameters"]["infos"]["iteration"] = trainer.updater.iteration
        config_session.set_metadata_modified_time()
        config_session.save_to(self.save_to + ".config")
# json.dump(config_session, open(self.save_to + ".config", "w"), indent=2,
# separators=(',', ': '))
        log.info("Saved trainer snapshot to file %s" % self.save_to)


def train_on_data_chainer(encdec, optimizer, training_data, output_files_dict,
                          src_indexer, tgt_indexer, eos_idx,
                          config_training,
                          stop_trigger=None,
                          test_data=None, dev_data=None, valid_data=None,
                          ):

    output_dir = config_training.training_management.save_prefix
    mb_size = config_training.training.mb_size
    nb_of_batch_to_sort = config_training.training.nb_batch_to_sort
    gpu = config_training.training_management.gpu
    report_every = config_training.training_management.report_every
    randomized = config_training.training.randomized_data
    reverse_src = config_training.training.reverse_src
    reverse_tgt = config_training.training.reverse_tgt
    do_not_save_data_for_resuming = config_training.training_management.no_resume
    noise_on_prev_word = config_training.training.noise_on_prev_word
    curiculum_training = config_training.training.curiculum_training
    use_previous_prediction = config_training.training.use_previous_prediction
    no_report_or_save = config_training.training_management.no_report_or_save
    use_memory_optimization = config_training.training_management.use_memory_optimization
    sample_every = config_training.training_management.sample_every
    use_reinf = config_training.training.use_reinf
    save_ckpt_every = config_training.training_management.save_ckpt_every
    trainer_snapshot = config_training.training_management.load_trainer_snapshot
    save_initial_model_to = config_training.training_management.save_initial_model_to
    reshuffle_every_epoch = config_training.training_management.reshuffle_every_epoch

    @chainer.training.make_extension()
    def sample_extension(trainer):
        encdec = trainer.updater.get_optimizer("main").target
        iterator = trainer.updater.get_iterator("main")
        mb_raw = iterator.peek()

        src_batch, tgt_batch, src_mask = make_batch_src_tgt(mb_raw, eos_idx=eos_idx, padding_idx=0, gpu=gpu, volatile="on", need_arg_sort=False)

        def s_unk_tag(num, utag):
            "S_UNK_%i" % utag

        def t_unk_tag(num, utag):
            "T_UNK_%i" % utag

        sample_once(encdec, src_batch, tgt_batch, src_mask, src_indexer, tgt_indexer, eos_idx,
                    max_nb=20,
                    s_unk_tag=s_unk_tag, t_unk_tag=t_unk_tag)

    iterator_training_data = LengthBasedSerialIterator(training_data, mb_size,
                                                       nb_of_batch_to_sort=nb_of_batch_to_sort,
                                                       sort_key=lambda x: len(x[0]),
                                                       repeat=True,
                                                       shuffle=reshuffle_every_epoch)

    def loss_func(src_batch, tgt_batch, src_mask):

        t0 = time.clock()
        (total_loss, total_nb_predictions), attn = encdec(src_batch, tgt_batch, src_mask, raw_loss_info=True,
                                                          noise_on_prev_word=noise_on_prev_word,
                                                          use_previous_prediction=use_previous_prediction,
                                                          mode="train")
        avg_loss = total_loss / total_nb_predictions

        t1 = time.clock()
        chainer.reporter.report({"forward_time": t1 - t0})

        chainer.reporter.report({"mb_loss": total_loss.data})
        chainer.reporter.report({"mb_nb_predictions": total_nb_predictions})
        chainer.reporter.report({"trg_loss": avg_loss.data})
        return avg_loss

    def convert_mb(mb_raw, device):
        return make_batch_src_tgt(mb_raw, eos_idx=eos_idx, padding_idx=0, gpu=device, volatile="off", need_arg_sort=False)

    updater = Updater(iterator_training_data, optimizer,
                      converter=convert_mb,
                      # iterator_training_data = chainer.iterators.SerialIterator(training_data, mb_size,
                      # repeat = True,
                      # shuffle = reshuffle_every_epoch)
                      device=gpu,
                      loss_func=loss_func,
                      need_to_convert_to_variables=False)

    trainer = chainer.training.Trainer(updater, stop_trigger, out=output_dir)
#     trainer.extend(chainer.training.extensions.LogReport(trigger=(10, 'iteration')))
#     trainer.extend(chainer.training.extensions.PrintReport(['epoch', 'iteration', 'trg_loss', "dev_loss", "dev_bleu"]),
#                    trigger = (1, "iteration"))

    if dev_data is not None and not no_report_or_save:
        dev_loss_extension = ComputeLossExtension(dev_data, eos_idx,
                                                  mb_size, gpu, reverse_src, reverse_tgt,
                                                  save_best_model_to=output_files_dict["model_best_loss"],
                                                  observation_name="dev_loss", config_training=config_training)
        trainer.extend(dev_loss_extension, trigger=(report_every, "iteration"))

        dev_bleu_extension = ComputeBleuExtension(dev_data, eos_idx, src_indexer, tgt_indexer,
                                                  output_files_dict["dev_translation_output"],
                                                  output_files_dict["dev_src_output"],
                                                  mb_size, gpu, reverse_src, reverse_tgt,
                                                  save_best_model_to=output_files_dict["model_best"],
                                                  observation_name="dev_bleu", config_training=config_training)

        trainer.extend(dev_bleu_extension, trigger=(report_every, "iteration"))

    if test_data is not None and not no_report_or_save:
        test_loss_extension = ComputeLossExtension(test_data, eos_idx,
                                                   mb_size, gpu, reverse_src, reverse_tgt,
                                                   observation_name="test_loss")
        trainer.extend(test_loss_extension, trigger=(report_every, "iteration"))

        test_bleu_extension = ComputeBleuExtension(test_data, eos_idx, src_indexer, tgt_indexer,
                                                   output_files_dict["test_translation_output"],
                                                   output_files_dict["test_src_output"],
                                                   mb_size, gpu, reverse_src, reverse_tgt,
                                                   observation_name="test_bleu")

        trainer.extend(test_bleu_extension, trigger=(report_every, "iteration"))

    if not no_report_or_save:
        trainer.extend(sample_extension, trigger=(sample_every, "iteration"))

        # trainer.extend(chainer.training.extensions.snapshot(), trigger = (save_ckpt_every, "iteration"))

        trainer.extend(CheckpontSavingExtension(output_files_dict["model_ckpt"], config_training), trigger=(save_ckpt_every, "iteration"))

        trainer.extend(SqliteLogExtension(db_path=output_files_dict["sqlite_db"]))

    trainer.extend(TrainingLossSummaryExtension(trigger=(report_every, "iteration")))

    if config_training.training_management.resume:
        if "model_parameters" not in config_training:
            log.error("cannot find model parameters in config file")
            raise ValueError(
                "Config file do not contain model_parameters section")
        if config_training.model_parameters.type == "snapshot":
            model_filename = config_training.model_parameters.filename
            log.info("resuming from trainer parameters %s" % model_filename)
            serializers.load_npz(model_filename, trainer)

    if trainer_snapshot is not None:
        log.info("loading trainer parameters from %s" % trainer_snapshot)
        serializers.load_npz(trainer_snapshot, trainer)

    try:
        if save_initial_model_to is not None:
            log.info("Saving initial parameters to %s" % save_initial_model_to)
            encdec = trainer.updater.get_optimizer("main").target
            serializers.save_npz(save_initial_model_to, encdec)

        trainer.run()
    except BaseException:
        if not no_report_or_save:
            final_snapshot_fn = output_files_dict["model_final"]
            log.info("Exception met. Trying to save current trainer state to file %s" % final_snapshot_fn)
            serializers.save_npz(final_snapshot_fn, trainer)
#             chainer.training.extensions.snapshot(filename = final_snapshot_fn)(trainer)
            config_session = config_training.copy(readonly=False)
            config_session.add_section("model_parameters", keep_at_bottom="metadata")
            config_session["model_parameters"]["filename"] = final_snapshot_fn
            config_session["model_parameters"]["type"] = "snapshot"
            config_session["model_parameters"]["description"] = "final"
            config_session["model_parameters"]["infos"] = argument_parsing_tools.OrderedNamespace()
            config_session["model_parameters"]["infos"]["iteration"] = trainer.updater.iteration
            config_session.set_metadata_modified_time()
            config_session.save_to(final_snapshot_fn + ".config")
# json.dump(config_session, open(final_snapshot_fn + ".config", "w"),
# indent=2, separators=(',', ': '))
            log.info("Saved trainer snapshot to file %s" % final_snapshot_fn)
        raise



""" Some utilities to leverage the argparse module and produce nicer config files."""

from collections import OrderedDict
import argparse
import json
import datetime
import sys
from nmt_chainer.utilities import versioning_tools


class OrderedNamespace(OrderedDict):
    """A class that act as a configuration dictionnary.
    Inherit from OrderedDict. When set to readonly mode, fields can be accessed by attributes:
        config["name"] == config.name
    """

    def __init__(self, *args, **kwargs):
        self.readonly = False
        super(OrderedNamespace, self).__init__(*args, **kwargs)

    def set_readonly(self):
        self.readonly = True
        for v in self.itervalues():
            if isinstance(v, OrderedNamespace):
                v.set_readonly()

    def __setitem__(self, *args, **kwargs):
        if self.readonly:
            raise KeyError("object in read-only mode")
        else:
            return OrderedDict.__setitem__(self, *args, **kwargs)

    def __getattr__(self, key):
        if self.readonly:
            try:
                return getattr(super(OrderedNamespace, self), key)
            except AttributeError:
                return self[key]
        else:
            return getattr(super(OrderedNamespace, self), key)

    def save_to(self, filename):
        json.dump(self, open(filename, "w"), indent=2, separators=(',', ': '))

    @classmethod
    def load_from(cls, filename):
        d = json.load(open(filename), object_pairs_hook=OrderedDict)
        cls.convert_to_ordered_namespace(d)
        return d

    @classmethod
    def convert_to_ordered_namespace(cls, ordered_dict):
        for v in ordered_dict.itervalues():
            if isinstance(v, OrderedDict):
                cls.convert_to_ordered_namespace(v)
        if not isinstance(ordered_dict, OrderedDict):
            raise ValueError()
        ordered_dict.__class__ = cls
        ordered_dict.readonly = False

    def update_recursive(self, other, valid_keys):
        if not isinstance(other, OrderedNamespace):
            raise ValueError()
        if self.readonly:
            raise ValueError()
        for key, val in other.iteritems():
            if isinstance(val, OrderedNamespace):
                if key in self:
                    if not (isinstance(self[key], OrderedNamespace)):
                        raise ValueError()
                else:
                    self[key] = OrderedNamespace()
                self[key].update_recursive(val, valid_keys)
            else:
                if key in self:
                    if (isinstance(self[key], OrderedNamespace)):
                        raise ValueError()
                if key in valid_keys:
                    self[key] = val

    def pretty_print(self, indent=4, discard_section=("metadata",)):
        for k, v in self.iteritems():
            if isinstance(v, OrderedNamespace) and k not in discard_section:
                print " " * indent, k, ":"
                v.pretty_print(indent=indent + 4, discard_section=())
            else:
                print " " * indent, k, v

    def copy(self, readonly=None):
        res = OrderedNamespace()
        for k, v in self.iteritems():
            if isinstance(v, OrderedNamespace):
                res[k] = v.copy(readonly)
            else:
                res[k] = v
        if readonly is None:
            res.readonly = self.readonly
        else:
            res.readonly = readonly
        return res

    def add_section(self, name, keep_at_bottom=None, overwrite=False):
        if self.readonly:
            raise ValueError()
        if name in self and not overwrite:
            return
        self[name] = OrderedNamespace()
        if keep_at_bottom is not None:
            metadata = self[keep_at_bottom]
            del self[keep_at_bottom]
            self[keep_at_bottom] = metadata

    def insert_section(self, name, content, even_if_readonly=False, keep_at_bottom=None, overwrite=False):
        if self.readonly and not even_if_readonly:
            raise ValueError()
        if name in self and not overwrite:
            raise ValueError()
        super(OrderedNamespace, self).__setitem__(name, content)
        if keep_at_bottom is not None:
            metadata = self[keep_at_bottom]
            del self[keep_at_bottom]
            super(OrderedNamespace, self).__setitem__(keep_at_bottom, metadata)

    def add_metadata_infos(self, version_num, overwrite=False):
        if self.readonly:
            raise ValueError()
        if "metadata" in self:
            if not overwrite:
                raise AssertionError()
        else:
            self["metadata"] = OrderedNamespace()

        self["metadata"]["config_version_num"] = version_num
        self["metadata"]["command_line"] = " ".join(sys.argv)
        self["metadata"]["knmt_version"] = versioning_tools.get_version_dict()
        self["metadata"]["creation_time"] = datetime.datetime.now().strftime("%I:%M%p %B %d, %Y")

    def set_metadata_modified_time(self):
        if self.readonly:
            raise ValueError()
        if "metadata" not in self:
            self["metadata"] = OrderedNamespace()
        self["metadata"]["modified_time"] = datetime.datetime.now().strftime("%I:%M%p %B %d, %Y")


class ParseOptionRecorder(object):
    """ A class whose main role is to remember the order in which argparse options have been defined, and to which subparser they belong.
    """

    def __init__(self, name=None, group_title_to_section=None, ignore_positional_arguments=set()):
        self.name = name
        self.argument_list = []
        self.group_title_to_section = group_title_to_section
        self.ignore_positional_arguments = ignore_positional_arguments

    def add_argument(self, argument_name, *args, **kwargs):
        positional = False
        if "dest" in kwargs:
            dest = kwargs["dest"]
            assert len(args) < 10
        elif len(args) >= 10:
            dest = args[9]
        elif argument_name.startswith("--"):
            dest = argument_name[2:]
        else:
            positional = True
            dest = argument_name
        if not (positional and dest in self.ignore_positional_arguments):
            self.argument_list.append(dest)

    def add_argument_group(self, title, desc=None):
        if self.group_title_to_section is not None:
            title = self.group_title_to_section[title]
        group = ParseOptionRecorder(title)
        self.argument_list.append(group)
        return group

    def convert_args_to_ordered_dict(self, args, args_is_namespace=True):
        result = OrderedNamespace()
        for arg in self.argument_list:
            if isinstance(arg, ParseOptionRecorder):
                if arg.name in result:
                    raise AssertionError
                result[arg.name] = arg.convert_args_to_ordered_dict(args, args_is_namespace=args_is_namespace)
            else:
                if arg in result:
                    raise AssertionError
                if arg in args:
                    if args_is_namespace:
                        result[arg] = getattr(args, arg)
                    else:
                        result[arg] = args[arg]
        return result


class ParserWithNoneDefaultAndNoGroup(object):
    """ A class whose main role is to help determine which arguments have been set on the command line."""

    def __init__(self):
        self.parser = argparse.ArgumentParser()

    def add_argument(self, *args, **kwargs):
        #         print " add_arg ", args, kwargs
        if len(args) >= 5:
            #             print "changing default args", args, kwargs
            assert "default" not in kwargs
            args[4] = None
#             print " -> ", args, kwargs
        elif "default" in kwargs:
            #             print "changing default", args, kwargs
            kwargs["default"] = None
#             print " -> ", args, kwargs
        self.parser.add_argument(*args, **kwargs)

    def add_argument_group(self, *args, **kwargs):
        return self

    def get_args_given(self, arglist):
        args_given_set = set()
        args, remaining_args = self.parser.parse_known_args(arglist)
        for argname in args.__dict__:
            assert argname not in args_given_set
            if getattr(args, argname) is not None:
                args_given_set.add(argname)
        return args_given_set


class ArgumentActionNotOverwriteWithNone(argparse.Action):
    """An argparse action that will not overwrite a dest that is already set to a non-None value.
    Useful for options that are set by multiple arguments"""

    def __call__(self, parser, namespace, values, option_string=None):
        if self.dest in namespace and getattr(
                namespace, self.dest) is not None and values is None:
            return
        setattr(namespace, self.dest, values)



import chainer
import cupy
import time
from cupy.cuda import Device

import threading
import Queue
import multiprocessing

import argparse
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("size", type = int)
parser.add_argument("gpu", nargs = "+", type = int)
args = parser.parse_args()

print "runnig exp with matrix size %i , on gpus %r"%(args.size, args.gpu)
t0 = time.clock()

x_list = {}
for n in args.gpu:
    print "creating data", n
    t1 = time.clock()
    with Device(n):
        t2 = time.clock()
        x = cupy.ones((args.size, args.size), dtype = cupy.float32)
        x_list[n] = x
    t3 = time.clock()
    print "time one creation:", t2-t1, t3-t2
 
for n in args.gpu:
    with Device(n):
        cupy.cuda.runtime.streamSynchronize(0)
t_sync1 = time.clock()
print "sync 1", t_sync1 - t0
 
 
def run_comp(n, appender):
    print "processing gpu", n
    t1b = time.clock()
    with Device(n):
        t2b = time.clock()
        print "using gpu", n
        x = x_list[n]
        t = x.dot(x)
        t += x.dot(x)
        print "middle_comp gpu", n
        t += x.dot(x)
        t += x.dot(x)
        t += x.dot(x)
        print "appending gpu", n
        appender(t)
    t3b = time.clock()
    print "time one comp:", t3b-t1b
    
# 
# 
# for u in theurls:
#     t = threading.Thread(target=get_url, args = (q,u))
#     t.daemon = True
#     t.start()


if 1:
    thread_class = threading.Thread
    queue_class = Queue.Queue
else:
    thread_class = multiprocessing.Process  
    queue_class = multiprocessing.Queue
    
q = queue_class()
res_list = []           
for n in args.gpu:
    t = thread_class(target = run_comp, args = (n, q.put))
#     t.daemon = True
    t.start()
res_list = []
while len(res_list) < len(args.gpu):
    res_list.append(q.get(block = True))

print "res_list filled"
    
# res_list = []           
# for n in args.gpu:
#     run_comp(n, appender = res_list.append)
    
for n in args.gpu:
    with Device(n):
        cupy.cuda.runtime.streamSynchronize(0)
        
t_sync2 = time.clock()
print "sync 2", t_sync2 - t_sync1
    
t4 = time.clock()   
print "computed all", len(res_list), t4-t0
print "summing across gpus:"
for i, n in enumerate(args.gpu[1:]):
    print "adding", i+1, n
    with Device(chainer.cuda.get_device(res_list[0])):
        t = res_list[i+1].copy()
        res_list[0] += t
t5a = time.clock()
print res_list[0]
t5 = time.clock()
print "sync time", t5-t5a
print "total_sum", t5-t4
print "total", t5-t0
    


#!/usr/bin/env python
"""rnn_cells.py: Wrappers around various RNN Cell types"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

from collections import Counter
import chainer
from chainer import cuda, Variable
from chainer import Link, Chain, ChainList
import chainer.functions as F
import chainer.links as L
from chainer import initializers

import logging
logging.basicConfig()
log = logging.getLogger("rnns:cells")
log.setLevel(logging.INFO)

# L.GRU = L.FastGRU
import nmt_chainer.utilities.faster_gru as faster_gru


class GRUCell(Chain):
    def __init__(self, in_size, out_size, init=None, inner_init=None, bias_init=None):
        log.info("Creating GRUCell(%i, %i)" % (in_size, out_size))
        super(GRUCell, self).__init__(
            gru=L.GRU(out_size, in_size, init=init, inner_init=inner_init, bias_init=bias_init),
        )
        self.add_param("initial_state", (1, out_size))
        self.initial_state.data[...] = self.xp.random.randn(out_size)
        self.out_size = out_size
        self.in_size = in_size

    def get_initial_states(self, mb_size):
        mb_initial_state = F.broadcast_to(F.reshape(
            self.initial_state, (1, self.out_size)), (mb_size, self.out_size))
        return (mb_initial_state,)

    def __call__(self, prev_states, x_in, mode="test"):
        assert mode in "test train".split()
        assert len(prev_states) == 1
        prev_state = prev_states[0]
        new_state = self.gru(prev_state, x_in)
        return (new_state,)

    def get_nb_states(self):
        return 1


class FastGRUCell(Chain):
    def __init__(self, in_size, out_size, init=None, bias_init=None):
        log.info("Creating GRUCell(%i, %i)" % (in_size, out_size))
        super(FastGRUCell, self).__init__(
            gru=faster_gru.GRU(out_size, in_size, init=init, bias_init=bias_init),
        )
        self.add_param("initial_state", (1, out_size))
        self.initial_state.data[...] = self.xp.random.randn(out_size)
        self.out_size = out_size
        self.in_size = in_size

    def get_initial_states(self, mb_size):
        mb_initial_state = F.broadcast_to(F.reshape(
            self.initial_state, (1, self.out_size)), (mb_size, self.out_size))
        return (mb_initial_state,)

    def __call__(self, prev_states, x_in, mode="test"):
        assert mode in "test train".split()
        assert len(prev_states) == 1
        prev_state = prev_states[0]
        new_state = self.gru(prev_state, x_in)
        return (new_state,)

    def get_nb_states(self):
        return 1


class LSTMCell(Chain):
    def __init__(self, in_size, out_size, lateral_init=None, upward_init=None, bias_init=None, forget_bias_init=None):
        log.info("Creating LSTMCell(%i, %i)" % (in_size, out_size))
        super(LSTMCell, self).__init__(
            lstm=L.StatelessLSTM(in_size, out_size,
                                 lateral_init=lateral_init, upward_init=upward_init,
                                 bias_init=bias_init, forget_bias_init=forget_bias_init)
        )
        self.add_param("initial_state", (1, out_size))
        self.initial_state.data[...] = self.xp.random.randn(out_size)
        self.add_persistent("initial_cell", self.xp.zeros((1, out_size), dtype=self.xp.float32))
        self.out_size = out_size
        self.in_size = in_size

    def get_initial_states(self, mb_size):
        mb_initial_state = F.broadcast_to(F.reshape(self.initial_state, (1, self.out_size)), (mb_size, self.out_size))
        mb_initial_cell = Variable(self.xp.broadcast_to(self.initial_cell, (mb_size, self.out_size)), volatile="auto")
        return (mb_initial_cell, mb_initial_state)

    def __call__(self, prev_states, x_in, mode="test"):
        assert mode in "test train".split()
        prev_cell, prev_state = prev_states
        new_cell, new_state = self.lstm(prev_cell, prev_state, x_in)
        return new_cell, new_state

    def get_nb_states(self):
        return 2  # state + cell


class GatedLSTMCell(Chain):
    def __init__(self, in_size, out_size, lateral_init=None, upward_init=None, bias_init=None, forget_bias_init=None):
        log.info("Creating GatedLSTMCell(%i, %i)" % (in_size, out_size))
        assert in_size == out_size

        super(GatedLSTMCell, self).__init__(
            lstm=L.StatelessLSTM(in_size, out_size, lateral_init=lateral_init, upward_init=upward_init, bias_init=bias_init, forget_bias_init=forget_bias_init),
            gate_w=L.Linear(in_size, in_size)
        )
        self.add_param("initial_state", (1, out_size))
        self.initial_state.data[...] = self.xp.random.randn(out_size)
        self.add_persistent("initial_cell", self.xp.zeros((1, out_size), dtype=self.xp.float32))
        self.add_persistent("initial_output", self.xp.zeros((1, out_size), dtype=self.xp.float32))
        self.out_size = out_size
        self.in_size = in_size

    def get_initial_states(self, mb_size):
        mb_initial_state = F.broadcast_to(F.reshape(self.initial_state, (1, self.out_size)), (mb_size, self.out_size))
        mb_initial_cell = Variable(self.xp.broadcast_to(self.initial_cell, (mb_size, self.out_size)), volatile="auto")
        mb_initial_output = Variable(self.xp.broadcast_to(self.initial_output, (mb_size, self.out_size)), volatile="auto")
        return (mb_initial_cell, mb_initial_state, mb_initial_output)

    def __call__(self, prev_states, x_in, mode="test"):
        assert mode in "test train".split()
        prev_cell, prev_state, old_output = prev_states
        new_cell, new_state = self.lstm(prev_cell, prev_state, x_in)

        passthrough_gate_state = F.sigmoid(self.gate_w(x_in))
        output = passthrough_gate_state * x_in + (1 - passthrough_gate_state) * new_state

        return new_cell, new_state, output

    def get_nb_states(self):
        return 3  # state + cell + gated_output


class StackedCell(ChainList):
    def __init__(self, in_size, out_size, cell_type=LSTMCell, nb_stacks=2,
                 dropout=0.5, residual_connection=False, no_dropout_on_input=False,
                 no_residual_connection_on_output=False, no_residual_connection_on_input=False,
                 init=None, inner_init=None, lateral_init=None, upward_init=None, bias_init=None, forget_bias_init=None):
        nb_stacks = nb_stacks or 2
        cell_type = cell_type or LSTMCell

        log.info("Creating StackedCell(%i, %i) X %i" % (in_size, out_size, nb_stacks))
        super(StackedCell, self).__init__()
        self.nb_of_states = []

        if cell_type in (LSTMCell, GatedLSTMCell):
            cell0 = cell_type(in_size, out_size, lateral_init=lateral_init, upward_init=upward_init, bias_init=bias_init, forget_bias_init=forget_bias_init)
        elif cell_type == GRUCell:
            cell0 = cell_type(in_size, out_size, init=init, inner_init=inner_init, bias_init=bias_init)
        elif cell_type == FastGRUCell:
            cell0 = cell_type(in_size, out_size, init=init, bias_init=bias_init)
        else:
            raise ValueError("Unsupported cell_type={0}".format(cell_type))
        self.add_link(cell0)
        self.nb_of_states.append(cell0.get_nb_states())

        for i in xrange(1, nb_stacks):
            if cell_type in (LSTMCell, GatedLSTMCell):
                cell = cell_type(out_size, out_size, lateral_init=lateral_init, upward_init=upward_init, bias_init=bias_init, forget_bias_init=forget_bias_init)
            elif cell_type == GRUCell:
                cell = cell_type(out_size, out_size, init=init, inner_init=inner_init, bias_init=bias_init)
            elif cell_type == FastGRUCell:
                cell = cell_type(out_size, out_size, init=init, bias_init=bias_init)
            else:
                raise ValueError("Unsupported cell_type={0}".format(cell_type))
            self.add_link(cell)
            self.nb_of_states.append(cell.get_nb_states())
        assert len(self) == nb_stacks

        self.dropout = dropout
        self.residual_connection = residual_connection
        self.no_residual_connection_on_output = no_residual_connection_on_output
        self.no_residual_connection_on_input = no_residual_connection_on_input
        self.no_dropout_on_input = no_dropout_on_input

    def get_initial_states(self, mb_size):
        res = []
        for i in xrange(len(self)):
            res += list(self[i].get_initial_states(mb_size))
        return tuple(res)

    def __call__(self, prev_states, x_in, mode="test"):
        assert mode in "test train".split()
        input_below = x_in
        states_cursor = 0
        res = []
        for i in xrange(len(self)):
            if self.dropout is not None and not (self.no_dropout_on_input and i == 0):
                input_below = F.dropout(input_below, ratio=self.dropout, train=(mode == "train"))
            new_states = self[i](prev_states[states_cursor:states_cursor + self.nb_of_states[i]], input_below,
                                 mode=mode)
            states_cursor += self.nb_of_states[i]

            if (self.residual_connection and
                not (i == len(self) - 1 and self.no_residual_connection_on_output) and
                    not (i == 0 and self.no_residual_connection_on_input)):
                input_below = new_states[-1] + input_below
            else:
                input_below = new_states[-1]

            res += list(new_states)
        return res


class NStepsCell(Chain):
    def __init__(self, in_size, out_size, nb_stacks, dropout, **kwds):
        super(NStepsCell, self).__init__(
            nstep_lstm=L.NStepLSTM(nb_stacks, in_size, out_size, dropout)
        )
        self.add_param("initial_state", (nb_stacks, 1, out_size))
        self.initial_state.data[...] = self.xp.random.randn(nb_stacks, 1, out_size)
        self.add_persistent("initial_cell", self.xp.zeros((nb_stacks, 1, out_size), dtype=self.xp.float32))

        self.nb_stacks = nb_stacks
        self.out_size = out_size
        self.in_size = in_size

    def get_initial_states(self, mb_size):
        mb_initial_state = F.broadcast_to(self.initial_state, (self.nb_stacks, mb_size, self.out_size))
        mb_initial_cell = Variable(self.xp.broadcast_to(self.initial_cell, (self.nb_stacks, mb_size, self.out_size)), volatile="auto")
        return (mb_initial_cell, mb_initial_state)

    def apply_to_seq(self, seq_list, mode="test"):
        assert mode in "test train".split()
        mb_size = len(seq_list)
        mb_initial_cell, mb_initial_state = self.get_initial_states(mb_size)
        return self.nstep_lstm(mb_initial_cell, mb_initial_state, seq_list, train=mode == "train")


# class DoubleGRU(Chain):
#     def __init__(self, H, I):
#         log.info("using double GRU")
#         self.H1 = H/2
#         self.H2 = H - self.H1
#         super(DoubleGRU, self).__init__(
#             gru1 = faster_gru.GRU(self.H1, I),
#             gru2 = faster_gru.GRU(self.H2, self.H1)
#         )
#
#     def __call__(self, prev_state, inpt):
#         prev_state1, prev_state2 = F.split_axis(prev_state, (self.H1,), axis = 1)
#
#         prev_state1 = self.gru1(prev_state1, inpt)
#         prev_state2 = self.gru2(prev_state2, prev_state1)
#
#         return F.concat((prev_state1, prev_state2), axis = 1)

cell_dict = {
    "lstm": LSTMCell,
    "dlstm": StackedCell,  # for backward compatibility
    "stack": StackedCell,
    "slow_gru": GRUCell,
    "gru": FastGRUCell,
    "glstm": GatedLSTMCell,
    "nsteps": NStepsCell
}

# has_dropout = set(["dlno_dropout_on_input = Falsestm"])


cell_description_keywords = {
    "dropout": float,
    "nb_stacks": int,
    "sub_cell_type": lambda k: cell_dict[k],
    "residual_connection": int,
    "no_dropout_on_input": int,
    "no_residual_connection_on_output": int,
    "no_residual_connection_on_input": int,
    "init_type": str,
    "inner_init_type": str,
    "lateral_init_type": str,
    "upward_init_type": str,
    "bias_init_type": str,
    "forget_bias_init_type": str,
    "init_scale": float,
    "inner_init_scale": float,
    "lateral_init_scale": float,
    "upward_init_scale": float,
    "bias_init_scale": float,
    "forget_bias_init_scale": float,
    "init_fillvalue": float,
    "inner_init_fillvalue": float,
    "lateral_init_fillvalue": float,
    "upward_init_fillvalue": float,
    "bias_init_fillvalue": float,
    "forget_bias_init_fillvalue": float
}


def create_cell_model_from_string(model_str):
    components = model_str.split(",")
    type_str = components[0]

    keywords = {}
    for comp in components[1:]:
        assert ":" in comp
        key, val = comp.split(":")
        if key in cell_description_keywords:
            keywords[key] = cell_description_keywords[key](val)
        else:
            raise ValueError("bad cell parameter: %s (possible parameters: %s)" %
                             (comp, " ".join(cell_description_keywords.keys())))

    return create_cell_model(type_str, **keywords)


def create_cell_model_from_config(config):
    type_str = config["cell_type"]
    keywords = dict((k, config[k]) for k in config if k != "cell_type")
    return create_cell_model(type_str, **keywords)


def create_initializer(init_type, scale=None, fillvalue=None):
    if init_type == 'identity':
        return initializers.Identity() if scale is None else initializers.Identity(scale=scale)
    if init_type == 'constant':
        return initializers.Constant(fillvalue)
    if init_type == 'zero':
        return initializers.Zero()
    if init_type == 'one':
        return initializers.One()
    if init_type == 'normal':
        return initializers.Normal() if scale is None else initializers.Normal(scale)
    if init_type == 'glorotNormal':
        return initializers.GlorotNormal() if scale is None else initializers.GlorotNormal(scale)
    if init_type == 'heNormal':
        return initializers.HeNormal() if scale is None else initializers.HeNormal(scale)
    if init_type == 'orthogonal':
        return initializers.Orthogonal(
            scale) if scale is None else initializers.Orthogonal(scale)
    if init_type == 'uniform':
        return initializers.Uniform(
            scale) if scale is None else initializers.Uniform(scale)
    if init_type == 'leCunUniform':
        return initializers.LeCunUniform(
            scale) if scale is None else initializers.LeCunUniform(scale)
    if init_type == 'glorotUniform':
        return initializers.GlorotUniform(
            scale) if scale is None else initializers.GlorotUniform(scale)
    if init_type == 'heUniform':
        return initializers.HeUniform(
            scale) if scale is None else initializers.HeUniform(scale)
    raise ValueError("Unknown initializer type: {0}".format(init_type))


def create_initializer_table(keywords):
    initializers = {}
    init_params = [k for k in keywords if 'init_' in k]
    init_kind = Counter(map((lambda str: str[:str.index('init') + 4]), init_params))
    for prefix in init_kind:
        init_type = None
        key = "{0}_type".format(prefix)
        if key in keywords:
            init_type = keywords[key]

        init_scale = None
        key = "{0}_scale".format(prefix)
        if key in keywords:
            init_scale = keywords[key]

        init_fillvalue = None
        key = "{0}_fillvalue".format(prefix)
        if key in keywords:
            init_fillvalue = keywords[key]

        initializer = create_initializer(init_type, init_scale, init_fillvalue)
        initializers[prefix] = initializer
    return initializers


def create_cell_model(type_str, **cell_parameters):
    initializers = create_initializer_table(cell_parameters)

    init = initializers['init'] if 'init' in initializers else None
    inner_init = initializers['inner_init'] if 'inner_init' in initializers else None
    lateral_init = initializers['lateral_init'] if 'lateral_init' in initializers else None
    upward_init = initializers['upward_init'] if 'upward_init' in initializers else None
    bias_init = initializers['bias_init'] if 'bias_init' in initializers else None
    forget_bias_init = initializers['forget_bias_init'] if 'forget_bias_init' in initializers else None

    if type_str not in cell_dict:
        raise ValueError("bad cell type: %s (possible types: %s)" %
                         (type_str, " ".join(cell_dict.keys())))
    cell_type = cell_dict[type_str]
    if type_str == "dlstm" or type_str == "stack" or type_str == "nsteps":
        def instantiate(in_size, out_size):
            # return cell_type(in_size, out_size, **cell_parameters)
            params = {}
            if 'sub_cell_type' in cell_parameters:
                params['cell_type'] = cell_parameters['sub_cell_type']
            if 'nb_stacks' in cell_parameters:
                params['nb_stacks'] = cell_parameters['nb_stacks']
            if 'dropout' in cell_parameters:
                params['dropout'] = cell_parameters['dropout']
            if 'residual_connection' in cell_parameters:
                params['residual_connection'] = cell_parameters['residual_connection']
            if 'no_dropout_on_input' in cell_parameters:
                params['no_dropout_on_input'] = cell_parameters['no_dropout_on_input']
            if 'no_residual_connection_on_output' in cell_parameters:
                params['no_residual_connection_on_output'] = cell_parameters['no_residual_connection_on_output']
            if 'no_residual_connection_on_input' in cell_parameters:
                params['no_residual_connection_on_input'] = cell_parameters['no_residual_connection_on_input']
            params['init'] = init
            params['inner_init'] = inner_init
            params['lateral_init'] = lateral_init
            params['upward_init'] = upward_init
            params['bias_init'] = bias_init
            params['forget_bias_init'] = forget_bias_init
            return cell_type(in_size, out_size, **params)

    else:
        def instantiate(in_size, out_size):
            if type_str in ("lstm", "glstm"):
                return cell_type(in_size, out_size, lateral_init=lateral_init, upward_init=upward_init, bias_init=bias_init, forget_bias_init=forget_bias_init)
            elif type_str == "slow_gru":
                return cell_type(in_size, out_size, init=init, inner_init=inner_init, bias_init=bias_init)
            elif type_str == "gru":
                return cell_type(in_size, out_size, init=init, bias_init=bias_init)
            else:
                raise ValueError("Unsupported cell_type={0}".format(cell_type))
    instantiate.meta_data_cell_type = cell_type
    return instantiate



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Author: Rico Sennrich

"""Use byte pair encoding (BPE) to learn a variable-length encoding of the vocabulary in a text.
Unlike the original BPE, it does not compress the plain text, but can be used to reduce the vocabulary
of a text to a configurable number of symbols, with only a small increase in the number of tokens.

Reference:
Rico Sennrich, Barry Haddow and Alexandra Birch (2016). Neural Machine Translation of Rare Words with Subword Units.
Proceedings of the 54th Annual Meeting of the Association for Computational Linguistics (ACL 2016). Berlin, Germany.
"""

# Dec. 2016: Edited by Raj Dabre

from __future__ import unicode_literals

import sys
import codecs
import re
import copy
import argparse
from collections import defaultdict, Counter

# hack for python2/3 compatibility
from io import open
argparse.open = open

# python 2/3 compatibility
# if sys.version_info < (3, 0):
#   sys.stderr = codecs.getwriter('UTF-8')(sys.stderr)
#   sys.stdout = codecs.getwriter('UTF-8')(sys.stdout)
#   sys.stdin = codecs.getreader('UTF-8')(sys.stdin)


def create_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="learn BPE-based word segmentation")

    parser.add_argument(
        '--input', '-i', type=argparse.FileType('r'), default=sys.stdin,
        metavar='PATH',
        help="Input text (default: standard input).")
    parser.add_argument(
        '--output', '-o', type=argparse.FileType('w'), default=sys.stdout,
        metavar='PATH',
        help="Output file for BPE codes (default: standard output)")
    parser.add_argument(
        '--symbols', '-s', type=int, default=10000,
        help="Create this many new symbols (each representing a character n-gram) (default: %(default)s))")
    parser.add_argument(
        '--min-frequency', type=int, default=2, metavar='FREQ',
        help='Stop if no symbol pair has frequency >= FREQ (default: %(default)s))')
    parser.add_argument(
        '--verbose', '-v', action="store_true",
        help="verbose mode.")

    return parser


def get_vocabulary(fobj):
    """Read text and return dictionary that encodes vocabulary
    """
    vocab = Counter()
    for line in fobj:
        for word in line.split():
            vocab[word] += 1
    return vocab


def get_vocabulary_and_totals(fobj):
    """Read text and return dictionary that encodes vocabulary along with total word count
    """
    total_words = 0
    total_lines = 0
    vocab = Counter()
    for line in fobj:
        total_lines += 1
        for word in line.split():
            vocab[word] += 1
            total_words += 1
    return vocab, total_words, total_lines


def update_pair_statistics(pair, changed, stats, indices):
    """Minimally update the indices and frequency of symbol pairs

    if we merge a pair of symbols, only pairs that overlap with occurrences
    of this pair are affected, and need to be updated.
    """
    stats[pair] = 0
    indices[pair] = defaultdict(int)
    first, second = pair
    new_pair = first + second
    for j, word, old_word, freq in changed:

        # find all instances of pair, and update frequency/indices around it
        i = 0
        while True:
            try:
                i = old_word.index(first, i)
            except ValueError:
                break
            if i < len(old_word) - 1 and old_word[i + 1] == second:
                if i:
                    prev = old_word[i - 1:i + 1]
                    stats[prev] -= freq
                    indices[prev][j] -= 1
                if i < len(old_word) - 2:
                    # don't double-count consecutive pairs
                    if old_word[i + 2] != first or i >= len(old_word) - 3 or old_word[i + 3] != second:
                        nex = old_word[i + 1:i + 3]
                        stats[nex] -= freq
                        indices[nex][j] -= 1
                i += 2
            else:
                i += 1

        i = 0
        while True:
            try:
                i = word.index(new_pair, i)
            except ValueError:
                break
            if i:
                prev = word[i - 1:i + 1]
                stats[prev] += freq
                indices[prev][j] += 1
            # don't double-count consecutive pairs
            if i < len(word) - 1 and word[i + 1] != new_pair:
                nex = word[i:i + 2]
                stats[nex] += freq
                indices[nex][j] += 1
            i += 1


def get_pair_statistics(vocab):
    """Count frequency of all symbol pairs, and create index"""

    # data structure of pair frequencies
    stats = defaultdict(int)

    # index from pairs to words
    indices = defaultdict(lambda: defaultdict(int))

    for i, (word, freq) in enumerate(vocab):
        prev_char = word[0]
        for char in word[1:]:
            stats[prev_char, char] += freq
            indices[prev_char, char][i] += 1
            prev_char = char

    return stats, indices


def replace_pair(pair, vocab, indices):
    """Replace all occurrences of a symbol pair ('A', 'B') with a new symbol 'AB'"""
    first, second = pair
    pair_str = ''.join(pair)
    pair_str = pair_str.replace('\\', '\\\\')
    changes = []
    pattern = re.compile(
        r'(?<!\S)' +
        re.escape(
            first +
            ' ' +
            second) +
        r'(?!\S)')
    if sys.version_info < (3, 0):
        iterator = indices[pair].iteritems()
    else:
        iterator = indices[pair].items()
    for j, freq in iterator:
        if freq < 1:
            continue
        word, freq = vocab[j]
        new_word = ' '.join(word)
        new_word = pattern.sub(pair_str, new_word)
        new_word = tuple(new_word.split())

        vocab[j] = (new_word, freq)
        changes.append((j, new_word, word, freq))

    return changes


def prune_stats(stats, big_stats, threshold):
    """Prune statistics dict for efficiency of max()

    The frequency of a symbol pair never increases, so pruning is generally safe
    (until we the most frequent pair is less frequent than a pair we previously pruned)
    big_stats keeps full statistics for when we need to access pruned items
    """
    for item, freq in list(stats.items()):
        if freq < threshold:
            del stats[item]
            if freq < 0:
                big_stats[item] += freq
            else:
                big_stats[item] = freq


def get_vocabulary_from_iterable(iterable):
    """Read text and return dictionary that encodes vocabulary
       iterable can be iterated and return sequences of "words"
    """
    vocab = Counter()
    for line in iterable:
        for word in line:
            vocab[word] += 1
    return vocab


def learn_bpe_from_sentence_iterable(iterable, output, symbols=10000, min_frequency=2, verbose=True):
    vocab = get_vocabulary_from_iterable(iterable)
    vocab = dict([(tuple(x) + ('</w>',), y) for (x, y) in vocab.items()])
    sorted_vocab = sorted(vocab.items(), key=lambda x: x[1], reverse=True)

    stats, indices = get_pair_statistics(sorted_vocab)
    big_stats = copy.deepcopy(stats)
    # threshold is inspired by Zipfian assumption, but should only affect speed
    threshold = max(stats.values()) / 10
    for i in range(symbols):
        if stats:
            most_frequent = max(stats, key=stats.get)

        # we probably missed the best pair because of pruning; go back to full
        # statistics
        if not stats or (i and stats[most_frequent] < threshold):
            prune_stats(stats, big_stats, threshold)
            stats = copy.deepcopy(big_stats)
            most_frequent = max(stats, key=stats.get)
            # threshold is inspired by Zipfian assumption, but should only affect speed
            threshold = stats[most_frequent] * i / (i + 10000.0)
            prune_stats(stats, big_stats, threshold)

        if stats[most_frequent] < min_frequency:
            sys.stderr.write(
                'no pair has frequency >= {0}. Stopping\n'.format(min_frequency))
            break

        if verbose:
            sys.stderr.write('pair {0}: {1} {2} -> {1}{2} (frequency {3})\n'.format(
                i, most_frequent[0], most_frequent[1], stats[most_frequent]))
        output.write('{0} {1}\n'.format(*most_frequent))
        changes = replace_pair(most_frequent, sorted_vocab, indices)
        update_pair_statistics(most_frequent, changes, stats, indices)
        stats[most_frequent] = 0
        if not i % 100:
            prune_stats(stats, big_stats, threshold)


if __name__ == '__main__':

    parser = create_parser()
    args = parser.parse_args()

    vocab = get_vocabulary(args.input)
    vocab = dict([(tuple(x) + ('</w>',), y) for (x, y) in vocab.items()])
    sorted_vocab = sorted(vocab.items(), key=lambda x: x[1], reverse=True)

    stats, indices = get_pair_statistics(sorted_vocab)
    big_stats = copy.deepcopy(stats)
    # threshold is inspired by Zipfian assumption, but should only affect speed
    threshold = max(stats.values()) / 10
    for i in range(args.symbols):
        if stats:
            most_frequent = max(stats, key=stats.get)

        # we probably missed the best pair because of pruning; go back to full
        # statistics
        if not stats or (i and stats[most_frequent] < threshold):
            prune_stats(stats, big_stats, threshold)
            stats = copy.deepcopy(big_stats)
            most_frequent = max(stats, key=stats.get)
            # threshold is inspired by Zipfian assumption, but should only affect speed
            threshold = stats[most_frequent] * i / (i + 10000.0)
            prune_stats(stats, big_stats, threshold)

        if stats[most_frequent] < args.min_frequency:
            sys.stderr.write(
                'no pair has frequency >= {0}. Stopping\n'.format(
                    args.min_frequency))
            break

        if args.verbose:
            sys.stderr.write('pair {0}: {1} {2} -> {1}{2} (frequency {3})\n'.format(
                i, most_frequent[0], most_frequent[1], stats[most_frequent]))
        args.output.write('{0} {1}\n'.format(*most_frequent))
        changes = replace_pair(most_frequent, sorted_vocab, indices)
        update_pair_statistics(most_frequent, changes, stats, indices)
        stats[most_frequent] = 0
        if not i % 100:
            prune_stats(stats, big_stats, threshold)



#!/usr/bin/env python
"""bleu_computer.py: compute BLEU score"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import argparse
import os

from collections import defaultdict
import math
import codecs
from itertools import izip


class BleuComputer(object):
    def __init__(self):
        self.ngrams_corrects = {1: 0, 2: 0, 3: 0, 4: 0}
        self.ngrams_total = {1: 0, 2: 0, 3: 0, 4: 0}
        self.total_length = 0
        self.ref_length = 0

    def copy(self):
        res = BleuComputer()
        res.ngrams_corrects = self.ngrams_corrects.copy()
        res.ngrams_total = self.ngrams_total.copy()
        res.total_length = self.total_length
        res.ref_length = self.ref_length
        return res

    def __repr__(self):
        res = []
        res.append("bleu:%f%%   " % (self.bleu() * 100))
        for n in xrange(1, 5):
            if self.ngrams_total[n] == 0:
                assert self.ngrams_corrects[n] == 0
                ratio_n = 1
            else:
                ratio_n = float(self.ngrams_corrects[n]) / self.ngrams_total[n]
            res.append("%i/%i[%f%%]" % (self.ngrams_corrects[n], self.ngrams_total[n], 100.0 * ratio_n))
        res.append("size of cand/ref: %i/%i[%f]" % (self.total_length, self.ref_length, float(self.total_length) / self.ref_length))
        return " ".join(res)

    __str__ = __repr__

    def bleu(self):
        if min(self.ngrams_corrects.values()) <= 0:
            return 0
        assert min(self.ngrams_total.values()) >= 0
        assert min(self.ngrams_total.values()) >= min(self.ngrams_corrects.values())

        log_brevity_penalty = min(0, 1.0 - float(self.ref_length) / self.total_length)
        log_average_precision = 0.25 * (
            sum(math.log(v) for v in self.ngrams_corrects.values()) -
            sum(math.log(v) for v in self.ngrams_total.values())
        )
        res = math.exp(log_brevity_penalty + log_average_precision)
        return res

    def bleu_plus_alpha(self, alpha=1.0):
        log_brevity_penalty = min(0, 1.0 - float(self.ref_length) / self.total_length)
        log_average_precision = 0.25 * (
            sum(math.log(v + alpha) for v in self.ngrams_corrects.values()) -
            sum(math.log(v + alpha) for v in self.ngrams_total.values())
        )
        res = math.exp(log_brevity_penalty + log_average_precision)
        return res

    def update(self, reference, translation):
        self.ref_length += len(reference)
        self.total_length += len(translation)
        for n in xrange(1, 5):
            reference_ngrams = defaultdict(int)
            translation_ngrams = defaultdict(int)
            for start in xrange(0, len(reference) - n + 1):
                ngram = tuple(reference[start: start + n])
                reference_ngrams[ngram] += 1
            for start in xrange(0, len(translation) - n + 1):
                ngram = tuple(translation[start: start + n])
#                 print ngram
                translation_ngrams[ngram] += 1
            for ngram, translation_freq in translation_ngrams.iteritems():
                reference_freq = reference_ngrams[ngram]
                self.ngrams_total[n] += translation_freq
                if ngram in reference_ngrams:
                    if reference_freq >= translation_freq:
                        self.ngrams_corrects[n] += translation_freq
                    else:
                        self.ngrams_corrects[n] += reference_freq

    def update_plus(self, diff):
        ngrams_corrects, ngrams_total, t_len, r_len = diff
        for n in xrange(1, 5):
            self.ngrams_corrects[n] += ngrams_corrects[n]
            self.ngrams_total[n] += ngrams_total[n]
        self.ref_length += r_len
        self.total_length += t_len

    def update_minus(self, diff):
        ngrams_corrects, ngrams_total, t_len, r_len = diff
        for n in xrange(1, 5):
            self.ngrams_corrects[n] -= ngrams_corrects[n]
            self.ngrams_total[n] -= ngrams_total[n]
            assert self.ngrams_corrects[n] >= 0
            assert self.ngrams_total[n] >= 0
        self.ref_length -= r_len
        self.total_length -= t_len

        assert self.total_length >= 0
        assert self.ref_length >= 0

    @staticmethod
    def compute_ngram_info(sentence):
        infos = defaultdict(int)
        for n in xrange(1, 5):
            for start in xrange(0, len(sentence) - n + 1):
                ngram = tuple(sentence[start: start + n])
                infos[ngram] += 1
        return infos, len(sentence)

    @staticmethod
    def compute_update_diff_from__infos(reference_info, translation_info):
        ngrams_corrects = {1: 0, 2: 0, 3: 0, 4: 0}
        ngrams_total = {1: 0, 2: 0, 3: 0, 4: 0}

        reference_ngrams, ref_len = reference_info
        translation_ngrams, t_len = translation_info

        for ngram, translation_freq in translation_ngrams.iteritems():
            n = len(ngram)
            reference_freq = reference_ngrams[ngram]
            ngrams_total[n] += translation_freq
            if ngram in reference_ngrams:
                if reference_freq >= translation_freq:
                    ngrams_corrects[n] += translation_freq
                else:
                    ngrams_corrects[n] += reference_freq
        return ngrams_corrects, ngrams_total, t_len, ref_len

    @staticmethod
    def compute_update_diff(reference, translation):
        ngrams_corrects = {1: 0, 2: 0, 3: 0, 4: 0}
        ngrams_total = {1: 0, 2: 0, 3: 0, 4: 0}

        for n in xrange(1, 5):
            reference_ngrams = defaultdict(int)
            translation_ngrams = defaultdict(int)
            for start in xrange(0, len(reference) - n + 1):
                ngram = tuple(reference[start: start + n])
                reference_ngrams[ngram] += 1
            for start in xrange(0, len(translation) - n + 1):
                ngram = tuple(translation[start: start + n])
#                 print ngram
                translation_ngrams[ngram] += 1
            for ngram, translation_freq in translation_ngrams.iteritems():
                reference_freq = reference_ngrams[ngram]
                ngrams_total[n] += translation_freq
                if ngram in reference_ngrams:
                    if reference_freq >= translation_freq:
                        ngrams_corrects[n] += translation_freq
                    else:
                        ngrams_corrects[n] += reference_freq
        return ngrams_corrects, ngrams_total, len(translation), len(reference)


def sample_bleu(ref_fn, trans_fn):
    import numpy
    ref_file = codecs.open(ref_fn, "r", encoding="utf8")
    trans_file = codecs.open(trans_fn, "r", encoding="utf8")
    infos_list = []
    for line_ref, line_trans in izip(ref_file, trans_file):
        r = line_ref.strip().split(" ")
        t = line_trans.strip().split(" ")
        infos_list.append(BleuComputer.compute_update_diff(r, t))
    nb_sentences = len(infos_list)
    while True:
        sample_idx = numpy.random.choice(nb_sentences, nb_sentences)
        bc = BleuComputer()
        for i in sample_idx:
            bc.update_plus(infos_list[i])
        yield bc


def get_bc_from_files(ref_fn, trans_fn):
    ref_file = codecs.open(ref_fn, "r", encoding="utf8")
    trans_file = codecs.open(trans_fn, "r", encoding="utf8")

    bc = BleuComputer()
    for line_ref, line_trans in izip(ref_file, trans_file):
        r = line_ref.strip().split(" ")
        t = line_trans.strip().split(" ")
        bc.update(r, t)
    return bc


def compute_confidence_interval_from_sampler(bleu_sampler, nb_resampling, confidence_interval=0.95):
    bleu_list = []
    for num_sample, bc in enumerate(bleu_sampler):
        if num_sample >= nb_resampling:
            break
        bleu_list.append(bc.bleu())
    bleu_list.sort()
    threshold_95 = int(nb_resampling * ((1.0 - confidence_interval) / 2.0))
    lower = bleu_list[threshold_95]
    higher = bleu_list[-threshold_95 - 1]
    sampled_mean = sum(bleu_list) / nb_resampling
#     print bleu_list[0], lower, higher, bleu_list[-1]
    return lower, higher, sampled_mean


def compute_pairwise_superiority_from_sampler_pair(bleu_sampler, other_bleu_sampler, nb_resampling):
    this = 0.0
    other = 0.0
    for num_sample, (bc, bc_other) in enumerate(
            izip(bleu_sampler, other_bleu_sampler)):
        if num_sample >= nb_resampling:
            break
        this_bleu = bc.bleu()
        other_bleu = bc_other.bleu()
        if this_bleu > other_bleu:
            this += 1
        if other_bleu > this_bleu:
            other += 1
    return this / nb_resampling, other / nb_resampling


def define_parser(parser):
    parser.add_argument("ref", help="reference file")
    parser.add_argument("translations", help="translations to be evaluated")
    parser.add_argument("--bootstrap", type=int, help="Do bootstrap resampling with that many resample")
    parser.add_argument("--other_translation", help="Another translation file for which we want to compare translation quality")


def do_bleu(args):
    if args.bootstrap is None:
        bc = get_bc_from_files(args.ref, args.translations)
        print bc
    else:
        bleu_sampler = sample_bleu(args.ref, args.translations)
        lower, higher, sampled_mean = compute_confidence_interval_from_sampler(bleu_sampler,
                                                                               nb_resampling=args.bootstrap,
                                                                               confidence_interval=0.95)

        bc = get_bc_from_files(args.ref, args.translations)
        print bc
        print "95%% confidence interval: %.02f - %.02f  (sampled mean:%.02f)" % (lower * 100, higher * 100, sampled_mean * 100)

        if args.other_translation is not None:
            other_bleu_sampler = sample_bleu(args.ref, args.other_translation)
            lower, higher, sampled_mean = compute_confidence_interval_from_sampler(other_bleu_sampler,
                                                                                   nb_resampling=args.bootstrap,
                                                                                   confidence_interval=0.95)
            bc = get_bc_from_files(args.ref, args.other_translation)
            print "\nOther translation:"
            print bc
            print "95%% confidence interval: %.02f - %.02f  (sampled mean:%.02f)" % (lower * 100, higher * 100, sampled_mean * 100)

            this_better, other_better = compute_pairwise_superiority_from_sampler_pair(
                bleu_sampler, other_bleu_sampler, nb_resampling=args.bootstrap)

            print
            print "better than other:", this_better * 100, "% of times"
            print "worse than other:", other_better * 100, "% of times"


def command_line():
    parser = argparse.ArgumentParser(description="Compute BLEU score")
    define_parser(parser)
    args = parser.parse_args()
    do_bleu(args)


if __name__ == "__main__":
    command_line()



"""replace_tgt_unk.py: Simple utility to replace target unknown words in nmt output"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import codecs
import itertools
import json
import unicodedata
import logging
logging.basicConfig()
log = logging.getLogger("rnns:replace_tgt")
log.setLevel(logging.INFO)


def replace_unk_from_string(translation_str, src_str, dic_fn, remove_unk, normalize_unicode_unk,
                            attempt_to_relocate_unk_source):

    dic = None
    if dic_fn is not None:
        dic = json.load(open(dic_fn))

    splitted_t = translation_str.strip().split(" ")
    splitted_s = src_str.strip().split(" ")
    new_t = []
    for p_w, w in enumerate(splitted_t):
        if w.startswith("#T_UNK_"):
            src_pos = int(w[7:-1])
            if src_pos >= len(splitted_s):
                log.warn("link to source out of bound (%i/%i line %i)" % (src_pos, len(splitted_s), num_line + 1))
                src_pos = len(splitted_s) - 1
            src_w = splitted_s[src_pos]

            if attempt_to_relocate_unk_source:
                if dic is not None and src_w in dic:
                    prec_w = splitted_t[p_w - 1] if p_w != 0 else None
#                     post_w = splitted_t[p_w  +1] if (p_w + 1) < len(splitted_t) else None
                    if prec_w is not None and dic[src_w] == prec_w and (src_pos + 1) < len(splitted_s):
                        log.info("retargeting unk  (%i/%i line %i)" % (src_pos, len(splitted_s), num_line + 1))
                        src_w = splitted_s[src_pos + 1]

#                 log.info("replacing %s %i"%(src_w, len(dic)))
            if dic is not None and src_w in dic:
                new_t.append(dic[src_w])
            else:
                #                     log.info("not found %s"%(src_w,))
                if not remove_unk:
                    if normalize_unicode_unk:
                        src_w = unicodedata.normalize("NFKD", src_w)
                    new_t.append(src_w)
        else:
            new_t.append(w)
    return " ".join(new_t) + "\n"


def replace_unk(translations, src_file, dest, dic_fn, remove_unk, normalize_unicode_unk,
                attempt_to_relocate_unk_source):

    ft = codecs.open(translations, encoding="utf8")
    fs = codecs.open(src_file, encoding="utf8")

    fd = codecs.open(dest, "w", encoding="utf8")

    dic = None
    if dic_fn is not None:
        dic = json.load(open(dic_fn))

    for num_line, (line_t, line_s) in enumerate(itertools.izip(ft, fs)):
        splitted_t = line_t.strip().split(" ")
        splitted_s = line_s.strip().split(" ")
        new_t = []
        for p_w, w in enumerate(splitted_t):
            if w.startswith("#T_UNK_"):
                src_pos = int(w[7:-1])
                if src_pos >= len(splitted_s):
                    log.warn("link to source out of bound (%i/%i line %i)" % (src_pos, len(splitted_s), num_line + 1))
                    src_pos = len(splitted_s) - 1
                src_w = splitted_s[src_pos]

                if attempt_to_relocate_unk_source:
                    if dic is not None and src_w in dic:
                        prec_w = splitted_t[p_w - 1] if p_w != 0 else None
    #                     post_w = splitted_t[p_w  +1] if (p_w + 1) < len(splitted_t) else None
                        if prec_w is not None and dic[src_w] == prec_w and (src_pos + 1) < len(splitted_s):
                            log.info("retargeting unk  (%i/%i line %i)" % (src_pos, len(splitted_s), num_line + 1))
                            src_w = splitted_s[src_pos + 1]

#                 log.info("replacing %s %i"%(src_w, len(dic)))
                if dic is not None and src_w in dic:
                    new_t.append(dic[src_w])
                else:
                    #                     log.info("not found %s"%(src_w,))
                    if not remove_unk:
                        if normalize_unicode_unk:
                            src_w = unicodedata.normalize("NFKD", src_w)
                        new_t.append(src_w)
            else:
                new_t.append(w)
        fd.write(" ".join(new_t) + "\n")


def define_parser(parser):
    parser.add_argument("translations")
    parser.add_argument("src_file")
    parser.add_argument("dest")
    parser.add_argument("--dic")
    parser.add_argument("--remove_unk", default=False, action="store_true")
    parser.add_argument("--normalize_unicode_unk", default=False, action="store_true")
    parser.add_argument("--attempt_to_relocate_unk_source", default=False, action="store_true")


def do_replace(args):
    replace_unk(args.translations, args.src_file, args.dest, args.dic, args.remove_unk,
                args.normalize_unicode_unk,
                args.attempt_to_relocate_unk_source)


def commandline():
    import argparse
    parser = argparse.ArgumentParser()
    define_parser(parser)
    args = parser.parse_args()
    do_replace(args)


if __name__ == '__main__':
    commandline()



import chainer
import chainer.function_hooks
from chainer import function
from chainer import cuda
from collections import defaultdict
import time
import numpy


def function_namer(function, in_data):
    in_shapes = []
    for in_elem in in_data:
        if hasattr(in_elem, "shape"):
            in_shapes.append("@%r" % (in_elem.shape,))
        elif isinstance(in_elem, (int, float, str)):
            in_shapes.append(in_elem)
        else:
            in_shapes.append("OTHER")
    if isinstance(function, chainer.functions.array.split_axis.SplitAxis):
        in_shapes.append(("s/a", function.indices_or_sections, function.axis))

    in_shapes = tuple(in_shapes)
    return (function.__class__, in_shapes)


class TimerElem(object):
    def __init__(self):
        self.fwd = 0
        self.bwd = 0
        self.total = 0
        self.nb_fwd = 0
        self.nb_bwd = 0

    def add_fwd(self, fwd):
        self.fwd += fwd
        self.total += fwd
        self.nb_fwd += 1

    def add_bwd(self, bwd):
        self.bwd += bwd
        self.total += bwd
        self.nb_bwd += 1

    def __repr__(self):
        return "<T:%f F[%i]:%f B[%i]:%f>" % (self.total, self.nb_fwd, self.fwd, self.nb_bwd, self.bwd)
    __str__ = __repr__


class MyTimerHook(function.FunctionHook):
    """Function hook for measuring elapsed time of functions.

    Attributes:
        call_history: List of measurement results. It consists of pairs of
            the function that calls this hook and the elapsed time
            the function consumes.
    """

    name = 'TimerHook'

    def __init__(self):
        self.call_times_per_classes = defaultdict(TimerElem)

    def __exit__(self, *_):
        print self
        self.print_sorted()
        print "total time:"
        print(self.total_time())
        super(MyTimerHook, self).__exit__(*_)

    def _preprocess(self):
        if self.xp == numpy:
            self.start = time.time()
        else:
            self.start = cuda.Event()
            self.stop = cuda.Event()
            self.start.record()

    def forward_preprocess(self, function, in_data):
        self.xp = cuda.get_array_module(*in_data)
        self._preprocess()

    def backward_preprocess(self, function, in_data, out_grad):
        self.xp = cuda.get_array_module(*(in_data + out_grad))
        self._preprocess()

    def _postprocess(self, function_repr, bwd=False):
        if self.xp == numpy:
            self.stop = time.time()
            elapsed_time = self.stop - self.start
        else:
            self.stop.record()
            self.stop.synchronize()
            # Note that `get_elapsed_time` returns result in milliseconds
            elapsed_time = cuda.cupy.cuda.get_elapsed_time(
                self.start, self.stop) / 1000.0
        if bwd:
            self.call_times_per_classes[function_repr].add_bwd(elapsed_time)
        else:
            self.call_times_per_classes[function_repr].add_fwd(elapsed_time)
#         self.call_history.append((function, elapsed_time))

    def forward_postprocess(self, function, in_data):
        xp = cuda.get_array_module(*in_data)
        assert xp == self.xp
        self._postprocess(function_namer(function, in_data))

    def backward_postprocess(self, function, in_data, out_grad):
        xp = cuda.get_array_module(*(in_data + out_grad))
        assert xp == self.xp
        self._postprocess(function_namer(function, in_data), bwd=True)

    def total_time(self):
        """Returns total elapsed time in seconds."""
        return sum(
            t.total for (
                _,
                t) in self.call_times_per_classes.iteritems())

    def print_sorted(self):
        for name, time in sorted(self.call_times_per_classes.items(), key=lambda x: x[1].total):
            print name, time



#!/usr/bin/env python
"""client.py: Client that can issue requests to KNMT Server."""
__author__ = "Frederic Bergeron"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "bergeron@pa.jst.jp"
__status__ = "Development"

import socket
import os.path
import re
from xml.sax.saxutils import escape


class Client:

    def __init__(self, server_ip, server_port):
        self.ip = server_ip
        self.port = server_port

    def query(self, sentence, article_id=1, beam_width=30, nb_steps=50, nb_steps_ratio=1.5,
              prob_space_combination=False, normalize_unicode_unk=True, remove_unk=False, attempt_to_relocate_unk_source=False,
              sentence_id=1):

        query = """<?xml version="1.0" encoding="utf-8"?>
<article id="{0}"
    beam_width="{1}"
    nb_steps="{2}"
    nb_steps_ratio="{3}"
    prob_space_combination="{4}"
    normalize_unicode_unk="{5}"
    remove_unk="{6}"
    attempt_to_relocate_unk_source="{7}">
    <sentence id="{8}">
        <i_sentence>{9}</i_sentence>
    </sentence>
</article>"""

        query = query.format(article_id, beam_width, nb_steps, nb_steps_ratio, prob_space_combination,
                             normalize_unicode_unk, remove_unk, attempt_to_relocate_unk_source, sentence_id, escape(sentence))
        s = socket.socket()
        s.connect((self.ip, self.port))
        s.send(query)

        try:
            resp = ''
            while True:
                data = s.recv(1024)
                resp += data
                if not data:
                    break
            return resp
        finally:
            s.close()



#!/usr/bin/env python

# from distutils.core import setup
from setuptools import setup, find_packages

# getting the version string
# as suggested in
# http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
import re
VERSIONFILE = "nmt_chainer/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


import subprocess
import os
import sys


def write_build_info():
    module_dir = os.path.dirname(os.path.realpath(__file__))
    try:
        f = open(os.path.join(module_dir, "nmt_chainer/_build.py"), "w")
    except OSError:
        print >>sys.stderr, "Warning: Could not create _build.py file"
        return

    try:
        git_hash = subprocess.check_output(['git', 'rev-parse', 'HEAD'],
                                           cwd=module_dir,
                                           stderr=subprocess.STDOUT).strip()
    except BaseException:
        git_hash = "**Could not retrieve git-hash"

    try:
        DEVNULL = open(os.devnull, 'wb')
        returncode = subprocess.call(['git', 'diff-index', '--quiet', 'HEAD', '--'],
                                     cwd=module_dir,
                                     stdout=DEVNULL,
                                     stderr=DEVNULL)
        if returncode == 0:
            git_dirty_status = "clean"
        elif returncode == 1:
            git_dirty_status = "dirty"
        else:
            git_dirty_status = "unknown"
    except BaseException:
        git_dirty_status = "unknown"

    try:
        git_diff = subprocess.check_output(['git', 'diff'],
                                           cwd=module_dir,
                                           stderr=subprocess.STDOUT)
    except BaseException:
        git_diff = '**Could not retrieve git-diff**'

    import json
    git_diff = json.dumps(git_diff)

    content = """# file created by setup.py
__build__ = "%s"

__dirty_status__ = "%s"
__git_diff__ = \"\"\"
%s
\"\"\"
""" % (git_hash, git_dirty_status, git_diff)

    f.write(content)


# classifiers
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

# create _build.py file
write_build_info()

# actual setup
setup(name='knmt',
      version=verstr,
      description='Implementation of RNNSearch and other Neural MT models in Chainer',
      author='Fabien Cromieres',
      author_email='fabien.cromieres@gmail.com',

      # packages=find_packages()
      packages=find_packages(),
      #       packages=['nmt_chainer'],
      url="https://github.com/fabiencro/knmt",
      #       scripts=["nmt_chainer/knmt.py"],

      scripts=["bin/knmt", "bin/knmt-debug", "bin/knmt-server"],

      #       entry_points= {
      #           'console_scripts': [
      #               'knmt = nmt_chainer.__main__:main'
      #           ]
      #                      },

      install_requires=['numpy>=1.10.0', 'chainer>=1.18.0', 'bokeh', 'plotly']
      )



#!/usr/bin/env python -O
# PYTHON_ARGCOMPLETE_OK

import argparse
import nmt_chainer.training_module.train_config as train
import nmt_chainer.translation.eval_config as eval_module
import nmt_chainer.dataprocessing.make_data_conf as make_data

import nmt_chainer.utilities.utils_command as utils_command

from nmt_chainer.utilities import versioning_tools
import sys

try:
    import argcomplete
except ImportError:
    argcomplete = None


def run_in_pdb(func, args):
    import pdb as pdb_module
    import sys
    import traceback
    pdb = pdb_module.Pdb()
    while True:
        try:
            pdb.runcall(func, args)
            if pdb._user_requested_quit:
                break
            print "The program finished and will be restarted"
        except pdb_module.Restart:
            print "Restarting with arguments:"
            print "\t" + " ".join(sys.argv[1:])
        except SystemExit:
            # In most cases SystemExit does not warrant a post-mortem session.
            print "The program exited via sys.exit(). Exit status: ",
            print sys.exc_info()[1]
        except SyntaxError:
            traceback.print_exc()
            sys.exit(1)
        except BaseException:
            traceback.print_exc()
            print "Uncaught exception. Entering post mortem debugging"
            print "Running 'cont' or 'step' will restart the program"
            t = sys.exc_info()[2]
            pdb.interaction(None, t)
            print "Post mortem debugger finished. The program will be restarted"


def main(arguments=None):
    # create the top-level parser
    parser = argparse.ArgumentParser(description="Kyoto-NMT: an Implementation of the RNNSearch model",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     prog="knmt")

    parser.add_argument("--run_in_pdb", default=False, action="store_true", help="run knmt in pdb (python debugger)")

    subparsers = parser.add_subparsers(dest="__subcommand_name")

    # create the parser for the "make_data" command
    parser_make_data = subparsers.add_parser('make_data', description="Prepare data for training.", help="Prepare data for training", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    make_data.define_parser(parser_make_data)

    # create the parser for the "train" command
    parser_train = subparsers.add_parser('train', description="Train a model.", help="Train a model", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    train.define_parser(parser_train)

    # create the parser for the "eval" command
    parser_eval = subparsers.add_parser('eval', description="Use a model.", help="Use a model", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    eval_module.define_parser(parser_eval)

    # create the parser for the "version" command
    parser_version = subparsers.add_parser('version', description="Get version infos.", help="Get version infos", formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser_utils = subparsers.add_parser('utils', description="Call a utility script.", help="Call a utility script", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    utils_command.define_parser(parser_utils)

#     import argcomplete
    if argcomplete is not None:
        argcomplete.autocomplete(parser)

    args = parser.parse_args(args=arguments)
    if arguments is not None:
        args.__original_argument_list = arguments
    else:
        args.__original_argument_list = sys.argv

    func = {"make_data": make_data.do_make_data,
            "train": train.do_train,
            "eval": eval_module.do_eval,
            "version": versioning_tools.main,
            "utils": utils_command.do_utils}[args.__subcommand_name]

    if args.run_in_pdb:
        run_in_pdb(func, args)
#         import pdb
#         pdb.runcall(func, args)
    else:
        try:
            func(args)
        except train.CommandLineValuesException as e:
            parser_train.error(e.args[0])
        except eval_module.CommandLineValuesException as e:
            parser_eval.error(e.args[0])
        except make_data.CommandLineValuesException as e:
            parser_make_data.error(e.args[0])


if __name__ == "__main__":
    main()
